import numpy as np
import scipy.constants as constants
from qetpy.core.didv._uncertainties_didv import get_dPdI_with_uncertainties, get_dVdI_with_uncertainties


__all__ = [
    "get_squid_noise_from_normal_noise", 
    "TESnoise",
]

def get_squid_noise_from_normal_noise(freqs=None, normal_noise=None,
                                      tload=0.1, tc=0.04, rload=0.01, rn=0.1, inductance=4e-7):
    """
    Takes UNFOLDED normal noise, a known temperature for the Rn and Rl, and
    generates the SQUID noise with the normal noise removed.
    
    If this is being used with the TESnoise class, you should us a normal noise
    that's from a trace that's an integer number of times longer than the noise
    trace being used to generate the noise being modeled by the TESnoise class, such
    that that class can get the frequencies it needs of the SQUID noise.
    
    Attributes
    ----------
    freqs: numpy array
        The frequencies at which the normal noise was measured
        
    normal_noise: numpy array
        The UNFOLDED normal TES noise, in units of Amps^2/Hz.
        
    tload: float
        The fit temperature of the load resistor in K.
        
    tc: float
        The tc of the device in K.
        
    rload : float
        The load resistance (rp + rsh) of the device in ohms.
        
    rn : float
        The normal resitance of the device in ohms.
        
    inductance : float
        The inductance of the device in Henries.
    """
    
    #rload = didv_result['biasparams']['rp'] + didv_result['biasparams']['rsh']
    #r0 = didv_result['biasparams']['r0']
    #inductance = didv_result['ssp_light']['vals']['L']
    r0 = rn
    
    omega = 2.0*np.pi*freqs
    dIdVnormal = 1.0/(rload + r0+1.0j*omega*inductance)
        
    s_vload = 2.0*constants.k*tload*rload * np.ones_like(freqs)
    
    s_iloadnormal = s_vload*np.abs(dIdVnormal)**2.0
   
    s_vtesnormal = 2.0*constants.k*tc*r0 * np.ones_like(freqs)
    
    s_itesnormal = s_vtesnormal*np.abs(dIdVnormal)**2.0
    
    
    s_itot = s_itesnormal + s_iloadnormal
    
    
    s_squid = normal_noise - s_itot
    return s_squid
    
                                      



class TESnoise:
    """
    Class for the simulation of the TES noise using the simple Irwin theory. Supports noise simulation for 
    in transition, superconducting, and normal.
    
    IMPORTANT NOTE:
    ALL NOISE MODELS WERE RECENTLY CHANGED TO BEING "UNFOLDED" BY ROGER, THEY WERE FOLDED IN SAM'S
    IMPLEMENTATION. THIS MEANS THAT THE NOISE COULD DIFFER BY A FACTOR OF 2 FROM WHAT YOU EXPECT!
    
    Attributes
    ----------
    freqs : float, array_like
        The frequencies for which we will calculate the noise simulation
        
    didv_result : 
        The dIdV fit result object generated using a QETpy dIdV fit
        
    tc : float
        The critical temperature of the TES in K
        
    tload : float
        The effective temperature of the load resistor in K
        
    tbath : float
        The bath temperature in K
        
    n : float
        The power-law dependence of the power flow to the heat bath
        
    lgc_ballistic : boolean
        Boolean flag that determines whether we use the ballistic (True) or
        diffusive limit when calculating TFN power noise
        
    f_tfn : float
        Function that estimates the noise suppression of the thermal fluctuation noise due
        to the difference in temperature between the bath and the TES. Supports the ballistic
        and diffusive limits, which is chosen via lgc_ballistic
        
    squid_noise_current : numpy array
        An array of the unfolded measured SQUID noise in units of amps^2 / Hz, measured from e.g. the
        normal noise with the normal TES noise subtracted out.
        
    """

    def __init__(self, freqs=None, didv_result=None, tc=0.040, tload=0.9, 
                 tbath=0.020, p0_manual=None, n=5.0, lgc_ballistic=True,
                 squid_noise_current=None, squid_noise_current_freqs=None,
                 lgc_diagnostics=True):
        """
        Initialization of the TES noise class.

        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation
        
        didv_result : 
            The dIdV fit result object generated using a QETpy dIdV fit
        
        tc : float, optional
            The critical temperature of the TES in K
            
        tload : float, optional
            The effective temperature of the load resistor in K
            
        tbath : float, optional
            The bath temperature in K
            
        p0_manual : float, optiona
            The TFN noise uses G, which is calculated from the tc and from the
            supplied bias power of the device. If 
            
        n : float, optional
            The power-law dependence of the power flow to the heat bath
            
        lgc_ballistic : boolean, optional
            Boolean flag that determines whether we use the ballistic (True) or
            diffusive limit when calculating TFN power noise
        
        squid_noise_current : numpy array
            An array of the unfolded measured SQUID noise in units of amps^2 / Hz, measured
            from e.g. the normal noise with the normal TES noise subtracted out. 
            Should be evaluated at the frequencies in squid_noise_current_freqs.
        
        squid_noise_current_freqs : numpy array
            The unfolded frequencies at which the SQUID noise was measured. Should include
            the frequencies which are in freqs (i.e. based on a SQUID noise measurement
            that's done from a trace that's an integer number of times the length of
            the trace which gives the frequencies at which the simulation is evaluated)
            
        lgc_diagnostics : bool, optional
            If True, prints out diagnostic messages.
            

        """
    
        if freqs is None:
            self.freqs = np.logspace(0, 5.5, 10000)
        else:
            self.freqs = freqs
            
        if didv_result is None:
            print("Must input a dIdV fit to use this noise simulation package class")
            
        self.didv_result = didv_result
            
        self.rload = didv_result['biasparams']['rp'] + didv_result['biasparams']['rsh']
        self.r0 = didv_result['biasparams']['r0']
        self.i0 = didv_result['biasparams']['i0']
        self.rshunt = didv_result['biasparams']['rsh']
        self.beta = didv_result['ssp_light']['vals']['beta']
        self.inductance = didv_result['ssp_light']['vals']['L']
        self.n = n
        self.tc = tc
        self.tload = tload
        self.tbath = tbath
        if p0_manual is None:
            self.G = self.n * didv_result['biasparams']['p0'] / self.tc
            if lgc_diagnostics:
                print("Automatically determining G")
                print("P0 = " + str(didv_result['biasparams']['p0']*1e15) + " fW")
                print("G = " +str(self.G) + " W/K") 
        else:
            self.G = self.n * p0_manual / self.tc
            if lgc_diagnostics:
                print("Manually setting G")
                print("P0 = " + str(p0_manual*1e15) + " fW")
                print("G = " +str(self.G) + " W/K")
        self.lgc_ballistic = lgc_ballistic
        
        self.squid_noise_current = squid_noise_current
        self.squid_noise_current_freqs = squid_noise_current_freqs
        
        if self.lgc_ballistic: # ballistic limit
            self.f_tfn = ((self.tbath/self.tc)**(self.n+1.0)+1.0)/2.0
        else:                  # diffusive limit
            self.f_tfn = self.n/(2.0*self.n+1.0) * ((self.tbath/self.tc)**(2.0*self.n+1.0)-1.0)/((self.tbath/self.tc)**(self.n)-1.0)
            
            
        if lgc_diagnostics:
            print("Calculating dVdI")
        self.dVdI, self.dVdI_err = get_dVdI_with_uncertainties(self.freqs, self.didv_result)
        self.dIdV = 1.0/self.dVdI
        if lgc_diagnostics:
            print("Calculating dPdI")
        self.dPdI, self.dPdI_err = get_dPdI_with_uncertainties(self.freqs, self.didv_result)
        self.dIdP = 1.0/self.dPdI
        if lgc_diagnostics:
            print("Done calculating dVdI and dPdI")
            
        self.lgc_diagnostics = lgc_diagnostics
        

    def s_vload(self, freqs=None):
        """
        The Johnson load voltage noise determined from the TES parameters, calculated for an
        UNFOLDED psd. This formula holds no matter where we are in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_vload : float, ndarray
            The UNFOLDED Johnson load voltage noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return 2.0*constants.k*self.tload*self.rload * np.ones_like(freqs)

    def s_iload(self, freqs=None):
        """
        The UNFOLDED Johnson load current noise determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_iload : float, ndarray
            The UNFOLDED Johnson load current noise at the specified frequencies
             
             
        """
        
        if freqs is None:
            freqs = self.freqs
            dVdI = self.dVdI
        else:
            dVdI, _ = get_dVdI_with_uncertainties(freqs, self.didv_result)
            
        return self.s_vload(freqs)*np.abs(dVdI)**-2.0

    def s_pload(self, freqs=None):
        """
        The UNFOLDED Johnson load power noise determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_pload : float, ndarray
            The UNFOLDED Johnson load power noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
            dPdI = self.dPdI
        else:
            dPdI, _ = get_dPdI_with_uncertainties(freqs, self.didv_result)
            
        return self.s_iload(freqs) * np.abs(dPdI)**2.0

    def s_vtes(self, freqs=None):
        """
        The UNFOLDED Johnson TES voltage noise determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_vtes : float, ndarray
            The Johnson TES voltage noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return 2.0*constants.k*self.tc*self.r0*(1.0+self.beta)**2.0 * np.ones(len(freqs), dtype='complex128')

    def s_ites(self, freqs=None):
        """
        The UNFOLDED Johnson TES current noise determined from the TES parameters for in transition. 
        This noise has both an electronic and thermal component.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_ites : float, ndarray
            The UNFOLDED Johnson TES current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
            dIdV = self.dIdV
            dIdP = self.dIdP
        else:
            dVdI, _ = get_dVdI_with_uncertainties(freqs, self.didv_result)
            dIdV = 1.0/dVdI
            dPdI, _ = get_dPdI_with_uncertainties(freqs, self.didv_result)
            dIdP = 1.0/dPdI
            
        return self.s_vtes(freqs)*np.abs(dIdV-self.i0*dIdP)**2.0

    def s_ptes(self, freqs=None):
        """
        The UNFOLDED Johnson TES power noise determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_ptes : float, ndarray
            The UNFOLDED Johnson TES power noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
            dPdI = self.dPdI
        else:
            dPdI, _ = get_dPdI_with_uncertainties(freqs, self.didv_result)
            
        return self.s_ites(freqs) * np.abs(dPdI)**2.0

    def s_ptfn(self, freqs=None):
        """
        The UNFOLDED thermal fluctuation noise in power determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_ptfn : float, ndarray
            The UNFOLDED thermal fluctuation noise in power at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return 2.0*constants.k*self.tc**2.0 * self.G * self.f_tfn * np.ones(len(freqs), dtype='complex128')

    def s_itfn(self, freqs=None):
        """
        The UNFOLDED thermal fluctuation noise in current determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_itfn : float, ndarray
            The UNFOLDED thermal fluctuation noise in current at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
            dIdP = self.dIdP
        else:
            dPdI, _ = get_dPdI_with_uncertainties(freqs, self.didv_result)
            dIdP = 1.0/dPdI

        return self.s_ptfn(freqs)*np.abs(dIdP)**2.0

    def s_isquid(self, freqs=None):
        """
        The UNFOLDED SQUID and downstream electronics current noise, derived from the real SQUID
        noise measured while normal.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_isquid : float, ndarray
            The SQUID and downstream electronics current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs

        sorted_indices = np.argsort(self.squid_noise_current_freqs)
        sorted_freqs = self.squid_noise_current_freqs[sorted_indices]
        sorted_noise = self.squid_noise_current[sorted_indices]
        
        squid_noise = np.interp(freqs, sorted_freqs, sorted_noise)

        return squid_noise
    

    def s_psquid(self, freqs=None):
        """
        The UNFOLDED SQUID and downstream electronics power noise, derived from the real SQUID
        noise measured while normal. This is only used for when the TES is in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_psquid : float, ndarray
            The UNFOLDED SQUID and downstream electronics power noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
            dPdI = self.dPdI
        else:
            dPdI, _ = get_dPdI_with_uncertainties(freqs, self.didv_result)
        
        return self.s_isquid(freqs=freqs) * dPdI**2

    def s_itot(self, freqs=None):
        """
        The total UNFOLDED current noise for the TES in transition. This is calculated by summing each of
        the current noise sources together. Units are [A^2/Hz].
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_itot : float, ndarray
            The total UNFOLDED current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_iload(freqs)+self.s_ites(freqs)+self.s_itfn(freqs)+self.s_isquid(freqs)

    def s_ptot(self, freqs=None): # total power noise [W^2/Hz]
        """
        The total UNFOLDED power noise for the TES in transition. This is calculated by summing each of
        the current noise sources together and using dIdP to convert to power noise. Units are [W^2/Hz].
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_ptot : float, ndarray
            The total UNFOLDED power noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
            dPdI = self.dPdI
        else:
            dPdI, _ = get_dPdI_with_uncertainties(freqs, self.didv_result)
            
        return self.s_itot(freqs) * np.abs(dPdI)**2.0
    
    def dIdVnormal(self, freqs=None):
        """
        The one-pole dIdV function determined from the TES parameters for when the TES
        is normal.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        dIdVnormal : float, ndarray
            The one-pole dIdV function for when the TES is normal.
                
        """
        
        if freqs is None:
            freqs = self.freqs
        omega = 2.0*np.pi*freqs
        return 1.0/(self.rload+self.r0+1.0j*omega*self.inductance)
    
    def s_iloadnormal(self, freqs=None):
        """
        The Johnson load current noise determined from the TES parameters for normal.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_iloadnormal : float, ndarray
            The Johnson load current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_vload(freqs)*np.abs(self.dIdVnormal(freqs))**2.0
    
    def s_vtesnormal(self, freqs=None):
        """
        The Johnson TES voltage noise determined from the TES parameters for normal.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_vtesnormal : float, ndarray
            The Johnson TES voltage noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return 2.0*constants.k*self.tc*self.r0 * np.ones_like(freqs)

    def s_itesnormal(self, freqs=None):
        """
        The Johnson TES current noise determined from the TES parameters for normal.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_itesnormal : float, ndarray
            The Johnson TES current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_vtesnormal(freqs)*np.abs(self.dIdVnormal(freqs))**2.0
    
    def s_itotnormal(self, freqs=None):
        """
        The total current noise for the TES when normal. This is calculated by summing each of
        the current noise sources together. Units are [A^2/Hz].
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_itotnormal : float, ndarray
            The total current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_iloadnormal(freqs)+self.s_itesnormal(freqs)+self.s_isquid(freqs)
    
    def dIdVsc(self, freqs=None):
        """
        The one-pole dIdV function determined from the TES parameters for when the TES
        is superconducting.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        dIdVsc : float, ndarray
            The one-pole dIdV function for when the TES is superconducting.
                
        """
        
        if freqs is None:
            freqs = self.freqs
        omega = 2.0*np.pi*freqs
        return 1.0/(self.rload+1.0j*omega*self.inductance)
    
    def s_iloadsc(self, freqs=None):
        """
        The Johnson load current noise determined from the TES parameters for superconducting.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_iloadsc : float, ndarray
            The Johnson load current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_vload(freqs)*np.abs(self.dIdVsc(freqs))**2.0
    
    def s_itotsc(self, freqs=None):
        """
        The total current noise for the TES when superconducting. This is calculated by summing each of
        the current noise sources together. Units are [A^2/Hz].
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_itotsc : float, ndarray
            The total current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_iloadsc(freqs)+self.s_isquid(freqs)




import numpy as np
import matplotlib.pyplot as plt


__all__ = [
    "plotnonlin",
]


def plotnonlin(OFnonlinOBJ, pulse, params, errors):
    """
    Diagnostic plotting of non-linear pulse fitting

    Parameters
    ----------
    OFnonlinOBJ: OFnonlin object
        The OFnonlin fit object to be plotted
    pulse: ndarray
        The raw trace to be fit
    params: tuple
        Tuple containing best fit paramters

    """

    if (OFnonlinOBJ.npolefit==4):
        A, B, C, tau_r,tau_f1, tau_f2, tau_f3, t0 = params
        A_err, B_err, C_err, tau_r_err, tau_f1_err, tau_f2_err, tau_f3_err, t0_err = errors
    elif (OFnonlinOBJ.npolefit==3):
        A, B, tau_r, tau_f1, tau_f2,t0 = params
        A_err, B_err, tau_r_err, tau_f1_err, tau_f2_err, t0_err = errors
    elif (OFnonlinOBJ.npolefit==2):
        A,tau_r, tau_f, t0 = params
        A_err, tau_r_err, tau_f_err, t0_err = errors
    else:
        A, tau_f, t0 = params
        A_err, tau_f_err, t0_err = errors
        tau_r = OFnonlinOBJ.taurise
        tau_r_err = 0.0

    if (OFnonlinOBJ.npolefit==4):
        variables = [A, B, C, tau_r, tau_f1, tau_f2, tau_f3, t0]
    elif (OFnonlinOBJ.npolefit==3):
        variables = [A, B, tau_r, tau_f1, tau_f2, t0]
    else:
        variables = [A, tau_r, tau_f, t0]
    ## get indices to define window ##
    t0ind = int(t0 * OFnonlinOBJ.fs) #location of timeoffset

    nmin = t0ind - int(5 * tau_r * OFnonlinOBJ.fs) # 5 risetimes before offset
    if (OFnonlinOBJ.npolefit==3 or OFnonlinOBJ.npolefit==4):
        nmax = t0ind + int(9 * tau_f1 * OFnonlinOBJ.fs) # 9 falltimes after offset
    else:
        nmax = t0ind + int(7 * tau_f * OFnonlinOBJ.fs) # 7 falltimes after offset

    nbaseline = int(OFnonlinOBJ.fs * t0) - 1000
    if nbaseline > 0:
        pulse = pulse - np.mean(pulse[:nbaseline])
    else:
        pulse = pulse - np.mean(pulse[nbaseline + 10000:])

    f = OFnonlinOBJ.freqs
    cf = f > 0
    f = f[cf]
    error = OFnonlinOBJ.error[cf]

    fig, axes = plt.subplots(2, 2, figsize = (12,8))
    if (OFnonlinOBJ.npolefit==4):
        fig.suptitle('Non-Linear Four Pole Fit', fontsize=18)
    elif (OFnonlinOBJ.npolefit==3):
        fig.suptitle('Non-Linear Three Pole Fit', fontsize=18)
    elif (OFnonlinOBJ.npolefit==2):
        fig.suptitle('Non-Linear Two Pole Fit', fontsize=18)
    elif (OFnonlinOBJ.npolefit==1):
        fig.suptitle('Non-Linear Two Pole Fit (Fixed Rise Time)', fontsize=18)

    axes[0][0].grid(linestyle='dashed')
    axes[0][0].set_title(f'Frequency Domain Trace')
    axes[0][0].set_xlabel(f'Frequency [Hz]')
    axes[0][0].set_ylabel('Amplitude [A/$\sqrt{\mathrm{Hz}}$]')
    axes[0][0].loglog(
        f, np.abs(OFnonlinOBJ.data[cf]), c='g', label='Pulse', alpha=0.75,
    )
    if (OFnonlinOBJ.npolefit==4):
        axes[0][0].loglog(
            f, np.abs(OFnonlinOBJ.fourpole(*variables))[cf], c='r', label='Fit',
        )
    elif (OFnonlinOBJ.npolefit==3):
        axes[0][0].loglog(
            f, np.abs(OFnonlinOBJ.threepole(*variables))[cf], c='r', label='Fit',
        )
    else:
        axes[0][0].loglog(
            f, np.abs(OFnonlinOBJ.twopole(*variables))[cf], c='r', label='Fit',
        )

    axes[0][0].loglog(f, error, c='b', label='$\sqrt{PSD}$', alpha=0.75)
    axes[0][0].tick_params(which='both', direction='in', right=True, top=True)

    axes[0][1].grid(linestyle = 'dashed')
    axes[0][1].set_title(f'Time Series Trace (Zoomed)')
    axes[0][1].set_xlabel(f'Time [ms]')
    axes[0][1].set_ylabel(f'Amplitude [Amps]')

    axes[0][1].plot(
        OFnonlinOBJ.time[nmin:nmax] * 1e3,
        pulse[nmin:nmax],
        c='g',
        label='Pulse',
        alpha=0.75,
    )

    if (OFnonlinOBJ.npolefit==4):
        axes[0][1].plot(
            OFnonlinOBJ.time[nmin:nmax] * 1e3,
            OFnonlinOBJ.fourpoletime(*variables)[nmin:nmax],
            c='r',
            label='Time Domain',
        )
    elif (OFnonlinOBJ.npolefit==3):
        axes[0][1].plot(
            OFnonlinOBJ.time[nmin:nmax] * 1e3,
            OFnonlinOBJ.threepoletime(*variables)[nmin:nmax],
            c='r',
            label='Time Domain',
        )
    else:
        axes[0][1].plot(
            OFnonlinOBJ.time[nmin:nmax] * 1e3,
            OFnonlinOBJ.twopoletime(*variables)[nmin:nmax],
            c='r',
            label='Time Domain',
        )
    axes[0][1].ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    axes[0][1].tick_params(which='both', direction='in', right=True, top=True)


    axes[1][0].grid(linestyle='dashed')
    axes[1][0].set_title(f'Time Series Trace (Full)')
    axes[1][0].set_xlabel(f'Time [ms]')
    axes[1][0].set_ylabel(f'Amplitude [Amps]')

    axes[1][0].plot(
        OFnonlinOBJ.time * 1e3, pulse, c='g', label='Pulse', alpha=0.75,
    )

    if (OFnonlinOBJ.npolefit==4):
        axes[1][0].plot(
            OFnonlinOBJ.time * 1e3,
            OFnonlinOBJ.fourpoletime(*variables),
            c='r',
            label='Time Domain',
        )
    elif (OFnonlinOBJ.npolefit==3):
        axes[1][0].plot(
            OFnonlinOBJ.time * 1e3,
            OFnonlinOBJ.threepoletime(*variables),
            c='r',
            label='Time Domain',
        )
    else:
        axes[1][0].plot(
            OFnonlinOBJ.time * 1e3,
            OFnonlinOBJ.twopoletime(*variables),
            c='r',
            label='Time Domain',
        )
    axes[1][0].ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    axes[1][0].tick_params(which='both', direction='in', right=True, top=True)

    axes[1][1].plot([], [], c='r', label='Best Fit')
    axes[1][1].plot([], [], c='g', label='Raw Data')
    axes[1][1].plot([], [], c='b', label='$\sqrt{PSD}$')

    for ii in range(len(params)):
        axes[1][1].plot([], [], linestyle=' ')

    if (OFnonlinOBJ.npolefit==4):
        labels = [
            f'A: ({A * 1e6:.4f} +\- {A_err * 1e6:.4f}) [$\mu$A]',
            f'B: ({B * 1e6:.4f} +\- {B_err * 1e6:.4f}) [$\mu$A]',
            f'C: ({C * 1e6:.4f} +\- {C_err * 1e6:.4f}) [$\mu$A]',
            f'τ, f1: ({tau_f1 * 1e6:.4f} +\- {tau_f1_err * 1e6:.4f}) [$\mu$s]',
            f'τ, f2: ({tau_f2 * 1e6:.4f} +\- {tau_f2_err * 1e6:.4f}) [$\mu$s]',
            f'τ, f3: ({tau_f3 * 1e6:.4f} +\- {tau_f3_err * 1e6:.4f}) [$\mu$s]',
            f'$t_0$: ({t0 * 1e3:.4f} +\- {t0_err * 1e3:.4f}) [ms]',
            f'τ$_r$: ({tau_r * 1e6:.4f} +\- {tau_r_err * 1e6:.4f}) [$\mu$s]',
        ]
    elif (OFnonlinOBJ.npolefit==3):
        labels = [
            f'A: ({A * 1e6:.4f} +\- {A_err * 1e6:.4f}) [$\mu$A]',
            f'B: ({B * 1e6:.4f} +\- {B_err * 1e6:.4f}) [$\mu$A]',
            f'τ, f1: ({tau_f1 * 1e6:.4f} +\- {tau_f1_err * 1e6:.4f}) [$\mu$s]',
            f'τ, f2: ({tau_f2 * 1e6:.4f} +\- {tau_f2_err * 1e6:.4f}) [$\mu$s]',
            f'$t_0$: ({t0 * 1e3:.4f} +\- {t0_err * 1e3:.4f}) [ms]',
            f'τ$_r$: ({tau_r * 1e6:.4f} +\- {tau_r_err * 1e6:.4f}) [$\mu$s]',
        ]
    else:
        labels = [
            f'A: ({A * 1e6:.4f} +\- {A_err * 1e6:.4f}) [$\mu$A]',
            f'τ$_f$: ({tau_f * 1e6:.4f} +\- {tau_f_err * 1e6:.4f}) [$\mu$s]',
            f'$t_0$: ({t0 * 1e3:.4f} +\- {t0_err * 1e3:.4f}) [ms]',
            f'τ$_r$: ({tau_r * 1e6:.4f} +\- {tau_r_err * 1e6:.4f}) [$\mu$s]',
        ]
    lines = axes[1][1].get_lines()
    legend1 = plt.legend(
        [lines[i] for i in range(3, 3 + len(params))],
        [labels[ii] for ii  in range(len(params))],
        loc=1,
    )
    legend2 = plt.legend(
        [lines[i] for i in range(0, 3)],
        ['Best Fit', 'Raw Data', '$\sqrt{PSD}$'],
        loc=2,
    )

    axes[1][1].add_artist(legend1)
    axes[1][1].add_artist(legend2)
    axes[1][1].axis('off')

    plt.tight_layout()
    plt.subplots_adjust(top=0.9)




import numpy as np
from math import ceil, floor
from scipy.optimize import least_squares
import matplotlib.pyplot as plt
from qetpy.utils import shift, interpolate_of, argmin_chisq
from qetpy.core import OFBase


__all__ = ['OF1x1',
           'get_time_offset_1x1']



class OF1x1:
    """
    Single trace /  single template optimal filter (1x1)
    calculations
    """

    def __init__(self, of_base=None,
                 channel='signal', template=None,
                 template_tag='default',
                 psd=None, sample_rate=None,
                 pretrigger_msec=None, pretrigger_samples=None,
                 coupling='AC', integralnorm=False,
                 verbose=True):
        
        """
        Initialize OF1x1

        Parameters
        ----------
        
        of_base : OFBase object, optional 
           OF base with pre-calculations
           Default: instantiate base class within OF1x1
        
        channel : str, optional
            channel name
        template : ndarray, optional
          template array used for OF calculation, can be 
        
        template_tag : str, optional 
           tamplate tag, default='default'
        
        psd : ndarray, optional
          psd array used for OF calculation, can be 
          None if already in of_base, otherwise required


        sample_rate : float, optional if of_base is not None
          The sample rate of the data being taken (in Hz)
          Only Required if of_base=None

        pretrigger_samples : int, optional if of_base is not None
            Number of pretrigger samples
            Default: use pretrigger_msec or if also None,
                     use 1/2 trace length

        pretrigger_msec : float, optional if of_base is not None
            Pretrigger length in ms (if  pretrigger_samples is None)
            Default: 1/2 trace length

        coupling : str, optional 
            String that determines if the zero frequency bin of the psd
            should be ignored (i.e. set to infinity) when calculating
            the optimum amplitude. If set to 'AC', then ths zero
            frequency bin is ignored. If set to anything else, then the
            zero frequency bin is kept.
            Default='AC'

        integralnorm : bool, optional
            If set to True, then  template will be normalized 
            to an integral of 1, and any optimum filters will
            instead return the optimum integral in units of Coulombs.
            If set to False, then the usual optimum filter amplitudes
            will be returned (in units of Amps).
            Default=False

        verbose : bool, optional 
            Display information
            Default=True


        Return
        ------
        
        """
    
        # verbose
        self._verbose = verbose

        # tag
        self._template_tag = template_tag

        # channel name
        self._channel_name = channel

        # Instantiate OF base (if not provided)
        self._of_base = of_base
        
        if of_base is None:

            # check parameters
            if sample_rate is None:
                raise ValueError('ERROR in OF1x1: sample rate required!')
                                     
            # instantiate
            self._of_base = OFBase(sample_rate, 
                                   verbose=verbose)
            
        # add template to base object
        if template is not None:

            if self._verbose:
                print('INFO: Adding template with tag "'
                      +  template_tag + '" to OF base object.')

            fs = self._of_base.sample_rate
                
            if pretrigger_msec is not None:
                pretrigger_samples = int(round(fs*pretrigger_msec/1000))
            elif pretrigger_samples is None:
                raise ValueError('ERROR in OF1x1: pretrigger '
                                 '(msec or samples) required!')
                
            self._of_base.add_template(channel,
                                       template,
                                       template_tag=template_tag,
                                       pretrigger_samples=pretrigger_samples,
                                       integralnorm=integralnorm)
        else:
            
            # check if template exist already
            tags =  self._of_base.template_tags(channel)
        
            if (tags is None
                or template_tag not in tags):

                raise ValueError(
                    f'ERROR: No template with tag "{template_tag}" '
                    f'for channel {channel} found in OF base object. '
                    f'Modify template tag or add template argument!')
                            
        # add noise to base object
        if psd is not None:

            if self._verbose:
                print('INFO: Adding noise PSD '
                      + 'to OF base object')
            
            self._of_base.set_psd(channel,
                                  psd,
                                  coupling=coupling)
            
        else:
            if self._of_base.psd(channel) is None:
                raise ValueError(f'ERROR: No psd found in OF base object.'
                                 f'for channel {channel}. Add psd argument!')
                
        #  template/noise pre-calculation
        if self._of_base.phi(channel, template_tag) is None:
            self._of_base.calc_phi(channel,
                                   template_tag=template_tag)

        # initialize fit results
        self.clear_results()
        

    def clear_results(self):
        """
        Clear fit results
        """

        self._of_amp_nodelay = None
        self._of_chi2_nodelay = None
        self._of_t0_nodelay = None
        self._of_chi2low_nodelay = None

        self._of_amp_withdelay = None
        self._of_chi2_withdelay = None
        self._of_t0_withdelay = None
        self._of_chi2low_withdelay = None

        # "no pulse" chi2
        self._of_chi2_nopulse = None

        # iterative
        self._of_amps_iterative = None
        self._of_chi2_iterative = None
        self._of_t0_iterative = None
        self._of_chi2low_iterative = None
        

    def get_result_nodelay(self):
        """
        Get OF no-delay results

        Parameters
        ----------
        none

        Return
        ------
  
        amp : float
            The optimum amplitude calculated for the trace (in Amps)
            with no time shifting allowed (or at the time specified by
            'shift_usec').

         t0 : float
            The time shift (=0 or shift_usec)

        chi2 : float
            The chi^2 value calculated from the optimum filter with no
            time shifting (or at the time shift specified by shift_usec)
            
        lowchi2_0 : float
            The low frequency chi^2 value (cut off at lowchi2_fcutoff) for the
            inputted values with no time shifting (or at the time shift 
            specified by shift_usec)
        
        """
        
        return (self._of_amp_nodelay,
                self._of_t0_nodelay,
                self._of_chi2_nodelay, 
                self._of_chi2low_nodelay)


    def get_result_withdelay(self):
        """
        Get OF with delay results
 
        Parameters
        ----------
        none

        Return
        ------

        amp : float
            The optimum amplitude calculated for the trace (in Amps).
        t0 : float
            The time shift calculated for the pulse (in s).
        chi2 : float
            The chi^2 value calculated from the optimum filter.
        lowchi2 : float
            The low frequency chi^2 value (cut off at lowchi2_fcutoff) for the
            inputted values
        """

        
        return (self._of_amp_withdelay,
                self._of_t0_withdelay,
                self._of_chi2_withdelay, 
                self._of_chi2low_withdelay)


    def get_chisq_nopulse(self):
        """
        Method to get 'no pulse' part of the chi2
        (independent of template)

        Parameters
        ---------
        None

        Return
        ------
        chi2_nopulse : float 
          
        """
        return self._of_chi2_nopulse
    
    
    def get_amplitude_resolution(self):
        """
        Method to return the energy resolution for the optimum filter.
        (resolution depends only on template and noise!) 
        
        Parameters
        ----------
        None


        Returns
        -------
        sigma : float
            The energy resolution of the optimum filter.

        """

        # get norm
        norm = self._of_base.weight(self._channel_name,
                                    template_tag=self._template_tag,
                                    squeeze_array=True)
        if norm is None:
            raise ValueError(f'ERROR: No "norm" found for '
                             f'channel {self._channel_name} and '
                             f'tag "{self._template_tag}" '
                             f'in OF base class!')
        

        sigma =  1.0 / np.sqrt(norm)
        
        return sigma

    
    def get_energy_resolution(self):
        """
        Deprecated method name: point to get_amplitude_resolution
        method
        """

        return self.get_amplitude_resolution()

    
    def get_time_resolution(self, lgc_nodelay=False):
        """
        Method to return the time resolution for the optimum filter.
        Resolution depends also on fitted amplitude (-> reset every events)
 
        Parameters
        ----------

        amp : float
          OF fitted amplitude
       

        Returns
        -------
        sigma : float
            The time resolution of the optimum filter.

        """


        # template FFT 
        template_fft = self._of_base.template_fft(
            self._channel_name,
            template_tag=self._template_tag,
            squeeze_array=True
        )
        
        if template_fft is None:
            raise ValueError(f'ERROR: No template FFT found for '
                             f'channel {self._channel_name} and '
                             f'tag "{self._template_tag}" '
                             f'in OF base class!')

        fft_freqs = self._of_base.fft_freqs()
        df = self._of_base.df()
        psd =  self._of_base.psd( self._channel_name)

        # check self._of_amp_nodelay = None
        amp = self._of_amp_withdelay
        if lgc_nodelay:
            amp =  self._of_amp_nodelay

        if amp is None:
            raise ValueError(f'ERROR: No fit done! '
                             f'Unable to calculate time resolution! ')
                
        sigma = 1.0 / np.sqrt(amp**2 * np.sum(
            (2*np.pi*fft_freqs)**2 * np.abs(template_fft)**2 / psd)
        )
        

        sigma = np.real(sigma)*np.sqrt(df)
        
        return sigma

            
    def calc(self, signal=None, 
             window_min_from_trig_usec=None,
             window_max_from_trig_usec=None,
             window_min_index=None,
             window_max_index=None,
             lowchi2_fcutoff=10000,
             lgc_outside_window=False,
             pulse_direction_constraint=0,
             interpolate_t0=False,
             lgc_fit_withdelay=True,
             lgc_fit_nodelay=True,
             lgc_plot=False):
        """
        Calculate OF with delay (and no delay if 
        lgc_fit_nodelay=True)

        Parameters
        ----------
        
        signal : ndarray, optional
          signal trace, can be None if already
          set in 'of_base
          
        window_min_from_trig_usec : float, optional
           OF filter window start in micro seconds from
           pre-trigger (can be negative if prior pre-trigger)

        window_max_from_trig_usec : float, optional
           OF filter window end in micro seconds from
           pre-trigger (can be negative if prior pre-trigger)


        window_min_index: int, optional
            OF filter window start in ADC samples 
        
        window_max_index: int, optional
            OF filter window end in ADC samples
            
        lowchi2_fcutoff : float, optional
            The frequency (in Hz) that we should cut off the chi^2 when
            calculating the low frequency chi^2. Default is 10 kHz.
            
        lgcoutsidewindow : bool, optional
            Boolean flag that is used to specify whether the Optimum
            Filter should look inside `nconstrain` or outside it. If
            False, the filter will minimize the chi^2 in the bins
            specified by `nconstrain`, which is the default behavior.
            If True, then it will minimize the chi^2 in the bins that
            do not contain the constrained window.


        pulse_direction_constraint : int, optional
            Sets a constraint on the direction of the fitted pulse.
            If 0, then no constraint on the pulse direction is set.
            If 1, then a positive pulse constraint is set for all fits.
            If -1, then a negative pulse constraint is set for all
            fits. If any other value, then a ValueError will be raised.
      
        interpolate_t0 : bool, optional
            If True, then a precomputed solution to the parabolic
            equation is used to find the interpolated time-of-best-fit.
            Default is False.

        lgc_fit_nodelay : bool, option
            calculation no-delay OF, default=True

        lgc_plot : bool, optional
            display diagnostic plot

        Return
        ------
        None

        """

        # clear results
        self.clear_results()
        
        # update signal and do preliminary
        # calculations
        if signal is not None:

            # clear
            self._of_base.clear_signal()

            # update
            self._of_base.update_signal(
                self._channel_name,
                signal,
                calc_fft=True
            )

        # calculate filtered signal
        if self._of_base.signal_filt(
            self._channel_name,
            template_tag=self._template_tag
        ) is None:
            
            self._of_base.calc_signal_filt(
                self._channel_name,
                template_tag=self._template_tag
            )

            self._of_base.calc_signal_filt_td(
                self._channel_name,
                template_tag=self._template_tag
            )

        # calc chisq
        self._calc_chisq_amp()
            
        # calc fit with delay
        if lgc_fit_withdelay:
            
            self._calc_fit_withdelay(
                window_min_from_trig_usec=window_min_from_trig_usec,
                window_max_from_trig_usec=window_max_from_trig_usec,
                window_min_index=window_min_index,
                window_max_index=window_max_index,
                lgc_outside_window=lgc_outside_window,
                pulse_direction_constraint=pulse_direction_constraint,
                interpolate_t0=interpolate_t0,
                lowchi2_fcutoff=lowchi2_fcutoff
            )

        
        # calc fit nodelay
        if lgc_fit_nodelay:
            
            self._calc_fit_nodelay(
                shift_usec=None,
                lowchi2_fcutoff=lowchi2_fcutoff,
                use_chisq_alltimes=True,
                lgc_plot=False)
                

        if lgc_plot:
            self.plot(lgc_plot_withdelay=True,
                      lgc_plot_nodelay=lgc_fit_nodelay)
                    
    def plot(self, lgc_plot_withdelay=True,
             lgc_plot_nodelay=True,
             figsize=(8, 5),
             xlim_msec=None):
        """
        Diagnostic plot

        Parameters
        ----------
        lgc_plot_withdelay : bool, optional
           If true, include OF with delay
           Default=True

        lgc_plot_nodelay : bool, optional
           If true, include OF no-delay
           Default=True

        figsize : tuple, optional
           figure size
           Default=(8, 5)

        xlim_msec : array like, optional
          min/max x-axis
        
        Return
        ------
        none
        
        """


        # check
        if lgc_plot_withdelay and self._of_amp_withdelay is None:
            print('WARNING: No fit (with delay) done. Unable to plot result!')
            return

        if lgc_plot_nodelay and self._of_amp_nodelay is None:
            print('WARNING: No fit (no delay) done. Unable to plot result!')
            return
        
        # signal
        signal = self._of_base.signal(self._channel_name, squeeze_array=True)
        template = self._of_base.template(self._channel_name,
                                          template_tag=self._template_tag,
                                          squeeze_array=True)
        
        fs = self._of_base.sample_rate
        nbins = len(signal)
        chi2 = self._of_chi2_withdelay/len(signal)
        
        # time axis
        xtime_ms = 1e3*np.arange(nbins)/fs
        
        # define figure abd plot
        fig, ax = plt.subplots(figsize=figsize)   
        ax.plot(xtime_ms, signal*1e6, label='Signal', color='blue', alpha=0.5)

        if lgc_plot_withdelay:
            chi2 = self._of_chi2_withdelay/len(signal)
            
            ax.plot(xtime_ms,
                    self._of_amp_withdelay*np.roll(
                        template,
                        int(self._of_t0_withdelay*fs)
                    )*1e6, 
                    label=(r'OF with delay, $\chi^2$'
                           + f'/Ndof={chi2:.2f}'),
                    color='red',
                    linestyle='dotted')

        if lgc_plot_nodelay:
            chi2 = self._of_chi2_nodelay/len(signal)
            ax.plot(xtime_ms,
                    self._of_amp_nodelay*np.roll(
                        template,
                        int(self._of_t0_nodelay*fs)
                    )*1e6, 
                    label=(r'OF no delay, $\chi^2$'
                           + f'/Ndof={chi2:.2f}'),
                    color='green',
                    linestyle='dotted')
       
        if xlim_msec is not None:
            ax.set_xlim(xlim_msec)
        ax.set_ylabel(r'Current [$\mu A$]')
        ax.set_xlabel('Time [ms]')
        ax.set_title(f'{self._channel_name} OF Results')
        lgd = ax.legend(loc='best')
        ax.tick_params(which='both', direction='in', right=True, top=True)
        ax.grid(linestyle='dotted')
        fig.tight_layout()
        

    def _calc_chisq0(self):
        """
        Calculate part of chi2 that doesn't depend
        on template (aka "no pulse chisq)
        
        Parameters
        ----------
        None


        Return
        ------
        None

        """

        # psd
        df = self._of_base.df()
        psd = self._of_base.psd(self._channel_name)
        
        # signal fft
        signal_fft = self._of_base.signal_fft(self._channel_name,
                                              squeeze_array=True)
        if signal_fft is None:
            raise ValueError(f'ERROR: No signal fft found for '
                             f'channel {self._channel_name} '
                             f'in OF base class!')
        
        
        # "no pulse chisq" (doesn't depend on template)
        self._of_chi2_nopulse = np.real(
            np.dot(signal_fft.conjugate()/psd,
                   signal_fft)
        )/df

                
        
    def _calc_chisq_amp(self):
        """
        Calculate chi2/amp for all times (rolled
        so that 0-delay is the pretrigger bin)

        Parameters
        ----------
        None

        Return
        ------
        None


        """

        # "no pulse chisq" (doesn't depend on template)
        self._calc_chisq0()

        # time dependent chisq + sum of the two

        # check if filtered signal (ifft) available
        # if not calculate
        signal_filt_td = self._of_base.signal_filt_td(
            self._channel_name,
            template_tag=self._template_tag,
            squeeze_array=True
        )
        
        if signal_filt_td is None:
            raise ValueError(f'ERROR: No filtered signal found for '
                             f'channel {self._channel_name} '
                             f'in OF base class!')
        
        # norm
        norm = self._of_base.weight(self._channel_name,
                                    template_tag=self._template_tag,
                                    squeeze_array=True)
        
        if norm is None:
            raise ValueError(f'ERROR: No "norm" found for '
                             f'channel {self._channel_name} '
                             f'in OF base class!')

        # pretrigger
        pretrigger_samples = self._of_base.nb_pretrigger_samples(
            self._channel_name,
            template_tag=self._template_tag
        )
            
        # initialize
        self._chisqs_alltimes_rolled = None
        self._amps_alltimes_rolled = None

        # amplitude
        amps_all = signal_filt_td/norm
        
        self._amps_alltimes_rolled = (
            np.roll(amps_all,
                    pretrigger_samples,
                    axis=-1)
        )

        
        # build chi2
        chisq_t0 = amps_all*signal_filt_td
        
        # total chisq
        chisq = self._of_chi2_nopulse - chisq_t0

        # shift so that 0 delay is at pre-trigger bin
        self._chisqs_alltimes_rolled = np.roll(
            chisq,
            pretrigger_samples,
            axis=-1
        )
        
     

        
    def _calc_fit_withdelay(self, 
                            window_min_from_trig_usec=None,
                            window_max_from_trig_usec=None,
                            window_min_index=None,
                            window_max_index=None,
                            lgc_outside_window=False,
                            pulse_direction_constraint=0,
                            interpolate_t0=False,
                            lowchi2_fcutoff=10000):
        """
        Function for calculating the optimum amplitude of a pulse in
        data with time delay. The OF window min/max can be specified
        either in usec from pretrigger or ADC samples. If no window,
        the all trace (unconstrained) is used.

        Parameters
        ----------

        window_min_from_trig_usec : float, optional
           OF filter window start in micro seconds from
           pre-trigger (can be negative if prior pre-trigger)


        window_max_from_trig_usec : float, optional
           OF filter window end in micro seconds from
           pre-trigger (can be negative if prior pre-trigger)


        window_min_index: int, optional
            OF filter window start in ADC samples

        window_max_index: int, optional
            OF filter window end in ADC samples

        lgcoutsidewindow : bool, optional
            Boolean flag that is used to specify whether the Optimum
            Filter should look inside `nconstrain` or outside it. If
            False, the filter will minimize the chi^2 in the bins
            specified by `nconstrain`, which is the default behavior.
            If True, then it will minimize the chi^2 in the bins that
            do not contain the constrained window.


        pulse_direction_constraint : int, optional
            Sets a constraint on the direction of the fitted pulse.
            If 0, then no constraint on the pulse direction is set.
            If 1, then a positive pulse constraint is set for all fits.
            If -1, then a negative pulse constraint is set for all
            fits. If any other value, then a ValueError will be raised.

        interpolate_t0 : bool, optional
            If True, then a precomputed solution to the parabolic
            equation is used to find the interpolated time-of-best-fit.
            Default is False.

        Returns
        -------
         None

        """

        # initialize
        amp = np.nan
        chisq = np.nan
        t0 = np.nan

        # sample rate and pretrigger
        fs = self._of_base.sample_rate
        nbins = self._of_base.nb_samples()
        pretrigger_samples = self._of_base.nb_pretrigger_samples(
            self._channel_name,
            template_tag=self._template_tag
        )

        # chisq and amp for all times
        chisqs_all = self._chisqs_alltimes_rolled
        amps_all = self._amps_alltimes_rolled

        # mask pulse direction
        constraint_mask = None
        if (pulse_direction_constraint==1
            or pulse_direction_constraint==-1):
            constraint_mask=(
                amps_all*pulse_direction_constraint>0
            )


        # find index minimum chisq within window
        window_min = None
        if window_min_from_trig_usec is not None:
            window_min = floor(pretrigger_samples
                               + window_min_from_trig_usec*fs*1e-6)
        elif window_min_index is not None:
            window_min = window_min_index

        if window_min is not None and window_min<0:
            window_min = 0

        window_max = None
        if window_max_from_trig_usec is not None:
            window_max = ceil(pretrigger_samples
                              + window_max_from_trig_usec*fs*1e-6)
        elif window_max_index is not None:
            window_max = window_max_index

        if window_max is not None and window_max>nbins:
            window_max = nbins

        if  window_min is not None:
             window_min = int(window_min)
             
        if  window_max is not None:
             window_max = int(window_max)

            
        bestind = argmin_chisq(
            chisqs_all,
            window_min=window_min,
            window_max=window_max,
            lgc_outside_window=lgc_outside_window,
            constraint_mask=constraint_mask
        )

        # extract chisq/amp (interpolate if requested)
        if np.isnan(bestind):
            amp = 0.0
            t0 = 0.0
            chisq = self._chisq_no_pulse
        elif interpolate_t0:
            amp, dt_interp, chisq = interpolate_of(
                amps_all, chisqs_all, bestind, 1/fs,
            )
            t0 = (bestind-pretrigger_samples)/fs + dt_interp
        else:
            amp = amps_all[bestind]
            t0 = (bestind-pretrigger_samples)/fs
            chisq = chisqs_all[bestind]
            
        # low frequency chisq
        lowchisq = self._get_chisq_lowfreq(
            amp, t0=t0,
            lowchi2_fcutoff=lowchi2_fcutoff
        )


        # store
        self._of_amp_withdelay = amp
        self._of_chi2_withdelay = chisq
        self._of_t0_withdelay = t0
        self._of_chi2low_withdelay = lowchisq

           

    def _calc_fit_nodelay(self,
                          shift_usec=None,
                          lowchi2_fcutoff=10000,
                          use_chisq_alltimes=True,
                          lgc_plot=False):
        """
        Calculate no-delay OF  (a shift
        from pretrigger can be added)

        Parameters
        ----------

        shift_usec : float, optional
          shift in micro seconds from pretrigger time
          default: no shift
                             
        lowchi2_fcutoff : float, optional
            The frequency (in Hz) that we should cut off the chi^2 when
            calculating the low frequency chi^2. Default is 10 kHz.
        
        use_chisq_alltimes : bool, optional
          use the chisq all times 
        
        lgc_plot : bool, optional
            display diagnostic plot


        Return
        ------
        None

        """

        # intialize
        amp_0 = np.nan
        chisq_0 = np.nan

        # sample rate and pretrigger. frequencies
        fs = self._of_base.sample_rate
        nbins = self._of_base.nb_samples()
        pretrigger_samples = self._of_base.nb_pretrigger_samples(
            self._channel_name,
            template_tag=self._template_tag
        )
        fft_freqs = self._of_base.fft_freqs()
        df = self._of_base.df()

        
        # shift
        t0_0 = 0
        t0_ind = pretrigger_samples
        if shift_usec is not None:
            t0_0 =  shift_usec*1e-6
            t0_ind += round(t0_0*fs)
            if t0_ind<0:
                t0_ind = 0
            elif t0_ind>nbins-1:
                t0_ind = nbins-1


        # use already calculated chisq/amp array
        if use_chisq_alltimes:

            if (self._amps_alltimes_rolled is None
                or self._chisqs_alltimes_rolled is None):
                self._calc_chisq_amp()

            amp_0 = self._amps_alltimes_rolled[t0_ind]
            chisq_0 = self._chisqs_alltimes_rolled[t0_ind]

        else:

            # get filtered signal
            signal_filt = self._of_base.signal_filt(
                self._channel_name,
                template_tag=self._template_tag,
                squeeze_array=True
            )
        
            if signal_filt is None:
                raise ValueError(f'ERROR: No filtered signal found for '
                                 f'channel {self._channel_name} '
                                 f'and tag {self._template_tag} '
                                 f'in OF base class!')


            # get norm
            norm = self._of_base.weight(self._channel_name,
                                        template_tag=self._template_tag,
                                        squeeze_array=True)
            
            # chisq no pulse 
            if self._of_chi2_nopulse is None:
                self._calc_chisq0()

            chi2_nopulse = self._of_chi2_nopulse.copy()
          
            # amplitude
            if shift_usec is not None:
                amp_0 = np.real(np.sum(
                    signal_filt*np.exp(2.0j*np.pi*t0_0*fft_freqs),
                    axis=-1,
                ))*df

            else:
                amp_0 = np.real(np.sum(
                    signal_filt, axis=-1
                ))*df

            # total chisq
            chisq_0 = chi2_nopulse - (amp_0**2)*norm


        # lowfreq chisq
        lowchisq_0 = self._get_chisq_lowfreq(
            amp_0, t0=t0_0,
            lowchi2_fcutoff=lowchi2_fcutoff
        )
        
        self._of_amp_nodelay = amp_0
        self._of_chi2_nodelay = chisq_0
        self._of_t0_nodelay = t0_0
        self._of_chi2low_nodelay = lowchisq_0

        if lgc_plot:
            self.plot(lgc_fit_nodelay=True)
            

    def _get_chisq_lowfreq(self, amp, t0=0,
                           lowchi2_fcutoff=10000):
        """
        Method for calculating the low frequency chi^2 of the optimum
        filter, given some cut off frequency.

        Parameters
        ----------
        amp : float
            The optimum amplitude calculated for the trace (in Amps).
        t0 : float, optional
            The time shift calculated for the pulse (in s).
            default: 0 (np shift)
        lowchi2_fcutoff : float, optional
            The frequency (in Hz) that we should cut off the chi^2 when
            calculating the low frequency chi^2. Default is 10 kHz.

   
        Returns
        -------
        chi2low : float
            The low frequency chi^2 value (cut off at lowchi2_fcutoff) for the
            inputted values.

        """

        # get template and signal FFT
        template_fft = self._of_base.template_fft(
            self._channel_name,
            template_tag=self._template_tag,
            squeeze_array=True
        )
        
        if template_fft is None:
            raise ValueError(f'ERROR: No template FFT found for '
                             f'channel {self._channel_name} and '
                             f'tag "{self._template_tag}" '
                             f'in OF base class!')

        # signal fft
        signal_fft = self._of_base.signal_fft(self._channel_name,
                                              squeeze_array=True)
        if signal_fft is None:
            raise ValueError(f'ERROR: No signal fft found for '
                             f'channel {self._channel_name} '
                             f'in OF base class!')

        # psd A^2/Hz
        psd = self._of_base.psd(self._channel_name)
        
        # sample rate and pretrigger. frequencies
        fs = self._of_base.sample_rate
        nbins = self._of_base.nb_samples()
        pretrigger_samples = self._of_base.nb_pretrigger_samples(
            self._channel_name,
            template_tag=self._template_tag
        )
        fft_freqs = self._of_base.fft_freqs()
        df = self._of_base.df()
        
        # calc chisq
        chi2tot = np.abs(
            signal_fft - amp * np.exp(-2.0j * np.pi * t0 * fft_freqs) * template_fft
        )**2 / psd
        
        # find low freq indices
        chi2inds = np.abs(fft_freqs) <= lowchi2_fcutoff
        ndof = np.sum(chi2inds)
            
        # sum
        chi2low = np.real(np.sum(chi2tot[chi2inds]))/df

        return chi2low

        
def get_time_offset_1x1(psd, template_1, template_2, fs=1.25e6, start_time=10e-3):
    """
    Calculates the offset between two different NxM templates, so that
    different templates will trigger at nominally the same time. There
    will still be some offset due to differences between the real
    event and the template, but the average difference should be zero.
    
    Parameters
    ----------
    
    psd : numpy array
        PSD used in the 1x1 OF.
        
    template_1 : numpy array
        Template for the first i.e. ''main trigger.'' The t0 output by 
        this function will tell you how far to offset template_2 to make
        it consistent with template_1.
        
    template_2 : 2x1xn numpy array
        Template for the second i.e. ''secondary trigger.'' The t0 output
        by this function will tell you how far to offset template_2 to make
        it consistent with template_1.
        
    fs : float, optional
        Sampling frequency (e.g. defaults to 1.25 MHz)
        
    start_time : float, optional
        The start time of the event, defaults to 10 ms.
        
    Return:
    -------
    
    t0 : float
        The time offset needed to be added to the start time of
        template_2 to make it consistent with template_1
    """
    
    channels= ['channel1', 'channel2']
    of = OF1x1(template=template_1, psd=psd, sample_rate=fs,
               pretrigger_samples=int(start_time*fs), 
               channel = 'ch1',
               verbose=False)
    
    of.calc(signal=template_2,
            window_min_from_trig_usec=-2000, 
            window_max_from_trig_usec=2000,
            lgc_fit_withdelay=True,
            lgc_fit_nodelay=False)
    amp, t0, chi2, lowchi2 = of.get_result_withdelay()
    return -t0



import numpy as np
from qetpy.utils import shift, interpolate_of, argmin_chisq
from qetpy.utils import convert_channel_name_to_list, convert_channel_list_to_name
from qetpy.core import OFBase
from numpy.linalg import pinv as pinv

__all__ = ['OFnxmx2']


class OFnxmx2:
    """
    N channels / M templates Optimal Fitler 
    (with 2 delays degree of freedom)
   
    """
    def __init__(self, of_base=None, channels=None,
                 templates=None, template_tag=None,
                 template_group_ids=None, csd=None, sample_rate=None,
                 pretrigger_msec=None, pretrigger_samples=None,
                 integralnorm=False, fit_window=None,
                 restrict_time_flag=True,
                 time_constraints_tag=None,
                 verbose=True):

        """
        Initialize OFnxmx2

        Parameters
        ----------

        of_base : OFBase object, optional
           OF base with pre-calculations
           Default: instantiate base class within OFnxmx2
  
        channels : str or list of string
          channels as ordered list or "|" separated string
          such as "channel1|channel2"
         
        templates : 3D ndarray [nchan, ntmp, nbins] (optional)
          multi-template array used for OF calculation, can be
          None if already in of_base, otherwise required
          
        template_tag : string (optional)
           tag associated with template matrix            

        template_group_ids : 2D array (optional)
          group idis used for multi-template array with each having 2 time degree of freedom
          used for OF calculation, can be None if already in of_base, otherwise required.
          eg for 4 template : one can have [0,1,0,0]-> meaning 1st, 3rd and fourth templat move together

        csd : ndarray, optional
          psd array used for OF calculation, can be
          None if already in of_base, otherwise required


        sample_rate : float, optional if of_base is not None
          The sample rate of the data being taken (in Hz)
          Only Required if of_base=None

        pretrigger_samples : int, optional if of_base is not None
            Number of pretrigger samples
            Default: use pretrigger_msec or if also None,
                     use 1/2 trace length

        pretrigger_msec : float, optional if of_base is not None
            Pretrigger length in ms (if  pretrigger_samples is None)
            Default: 1/2 trace length

        integralnorm : bool, optional
            If set to True, then  template will be normalized
            to an integral of 1, and any optimum filters will
            instead return the optimum integral in units of Coulombs.
            If set to False, then the usual optimum filter amplitudes
            will be returned (in units of Amps).
            Default=False

        channel_name : str, optional
            channel name

        fit_window :  used for preparing the time window for nxmx2 filter, 
                      if not specified we will use
                      the available bins size

             time_combinations1 = np.arange(int(fit_window[0][0]), int(fit_window[0][1]))
             time_combinations2 = np.arange(int(fit_window[1][0]), int(fit_window[1][1]))

        restrict_time_flag : boolean, optional

        time_constraints_tag : str (optional)
          tag associated to template_group_ids,  fit_window, restrict_time_flag

        verbose : bool, optional
            Display information
            Default=True


        Return
        ------

        """

        # verbose
        self._verbose = verbose
        
        # channels
        if channels is None:
            raise ValueError('ERROR: "channels" argument required!')
        
        self._channel_name = convert_channel_list_to_name(channels)
        self._channel_list = convert_channel_name_to_list(channels)
        self._nchans = len(self._channel_list)
        
        # initialize number of templates (per channel)
        self._ntmps = None

        # check templates argument
        if templates is None:
            
            # OF base required
            if of_base is None:
                raise ValueError('ERROR: Either "of_base" or '
                                 '"templates" argument required!')

            # templates tag required 
            if template_tag is None:
                raise ValueError('ERROR: "template_tag" required '
                                 'if "of_base" provided and "templates" '
                                 'argument is None')
        
        else:

            # check numpy array
            if not isinstance(templates, np.ndarray):
                raise ValueError('ERROR: Expecting "templates" to be '
                                 'a numpy array')
        
            if templates.ndim == 1:
                templates = templates[np.newaxis, np.newaxis, :]  # 1 channel, 1 template
            elif templates.ndim == 2:
                templates = templates[np.newaxis, :, :]            # 1 channel, multiple templates
            elif templates.ndim != 3:
                raise ValueError('ERROR: "templates" input must be 1D, 2D, or 3D')    
            
            if templates.shape[0] != self._nchans:
                raise ValueError(f'ERROR: Expecting "templates" to have '
                                 f'shape[0]={self._nchans}!')
                        
            # check template tags
            if template_tag is None:
                template_tag = 'default'

        # Instantiate OF base (if None) and add templates
        self._of_base = of_base

        # instantiate OF base if None
        if of_base is None:
            
            # check required parameters
            if sample_rate is None:
                raise ValueError('ERROR in OFnxm: sample rate required!')

            if (pretrigger_msec is None
                and pretrigger_samples is None):
                raise ValueError('ERROR in OFnxm: '
                                 'pretrigger (msec or samples) required!')

            # check time cosntraints
            if (template_group_ids is None
                or fit_window is None):
                raise ValueError('ERROR: "template_group_ids" and '
                                 '"fit_window" required!')
            
            # instantiate
            self._of_base = OFBase(sample_rate, verbose=verbose)

            
        # add template to base object if templates not None
        if templates is not None:
        
            fs = self._of_base.sample_rate
            if pretrigger_msec is not None:
                pretrigger_samples = int(round(fs*pretrigger_msec/1000))
            # verbose
            print(f'INFO: Adding templates with shape={templates.shape} '
                  'to OF base object!')
                                    
            # add to OF
            self._of_base.add_template(
                self._channel_name, templates, template_tag,
                pretrigger_samples=pretrigger_samples,
                integralnorm=integralnorm,
                overwrite=True
            )
                        
        else:

            templates = self._of_base.template(
                self._channel_name,
                template_tag=template_tag
            )

            if templates is None:
                raise ValueError(
                    f'ERROR: No template with tag '
                    f'"{template_tag}" found for channel '
                    f'{self._channel_name} in OF base!'
                )
                    
        # save template tags
        self._template_tag = template_tag
        self._ntmps = templates.shape[1]
        
        # sample rate
        self._fs = self._of_base.sample_rate
        
        # number of samples
        self._nb_samples = self._of_base.nb_samples()
        self._nb_pretrigger_samples = self._of_base.nb_pretrigger_samples(
            self._channel_name, self._template_tag
        )

        #  add time constraint to base object if templates not None
        if (template_group_ids is not None
            and fit_window is not None):

            if time_constraints_tag is None:
                time_constraints_tag = 'default'
                
            self._of_base.set_time_constraints(
                self._channel_name,
                template_group_ids=template_group_ids,
                fit_window=fit_window,
                restrict_time_flag=restrict_time_flag,
                time_constraints_tag=time_constraints_tag
            )  
            
        elif time_constraints_tag is None:
            raise ValueError(
                'ERROR: "time_constraints_tag" required if no input '
                '"template_group_ids" and "fit_window"'
            )
        
        
        # get time combinations
        time_combinations =  self._of_base.time_combinations(
            self._channel_name,
            time_constraints_tag=time_constraints_tag)

        if time_combinations is None:
            raise ValueError(
                'ERROR: time constraints missing from OF base! '
                'Arguments "template_group_ids" and "fit_window" '
                'need to be provided')

        self._time_constraints_tag = time_constraints_tag
        self._time_combinations = time_combinations
        self._template_group_ids = (
            self._of_base.template_group_ids(
                self._channel_name,
                time_constraints_tag=time_constraints_tag)
        )

        # add noise to base object
        if csd is not None:
            
            if self._verbose:
                print('INFO: Adding noise CSD to OF base object')
                
            self._of_base.set_csd(channels=self._channel_name, csd=csd)
            
        elif self._of_base.csd(channels=self._channel_name) is None:
            raise ValueError('ERROR: No csd found in OF base object.'
                             ' Add csd argument!')
                  
        #  template/noise pre-calculation
        #   aka "p" matrix
        
        
        # initialize
        self.clear()
        
        # check if p matrix calculated
        if  self._of_base.p_matrix(
                self._channel_name,
                template_tag=template_tag,
                time_constraints_tag=time_constraints_tag) is None:
        
            self._of_base.calc_p_matrix(
                self._channel_name,
                template_tag=template_tag,
                time_constraints_tag=time_constraints_tag
            )


        # save
        self._p_matrix = self._of_base.p_matrix(
            self._channel_name,
            template_tag=template_tag,
            time_constraints_tag=time_constraints_tag)
    
        self._p_matrix_inv = self._of_base.p_matrix_inv(
            self._channel_name,
            template_tag=template_tag,
            time_constraints_tag=time_constraints_tag
        )

                

    def clear(self):
        """
        clear all signal data
        """
        # initialize
        self._q_vector  = None
        self._amps_allt = None
        self._chi2_allt = None

        self._of_amps = None
        self._of_chi2 = None
        self._of_t0 = None
        self._index_first_pulse = None
        self._index_second_pulse =  None
        self._of_chi2_per_DOF = None
        
        
    def calc(self, signal=None, polarity_constraint=False):
        """
        FIXME
        docstrings need to be added with dimensions
        update_signal needs to be called ? from of_base
        """
        
        # update signal and do preliminary (signal) calculations
        # (like fft signal, filtered signals, signal matrix ...

        # clear internal (signal) data
        self.clear()
        
        # update signal and do preliminary (signal) calculations
        # (like fft signal, filtered signal, signal matrix ...
        if signal is not None:
            
            # check if correct size
            if (not isinstance(signal, np.ndarray)
                or signal.shape[0] != self._nchans):
                raise ValueError('ERROR: Wrong number of channels '
                                 'in "signal" array!')
            
            self._of_base.clear_signal()

            # update signal
            self._of_base.update_signal(
                self._channel_name,
                signal,
                calc_fft=True)

            # calculate signal filter
            self._of_base.calc_signal_filt(
                self._channel_name,
                template_tag=self._template_tag
            )
            
            # calc signal filter ifft
            self._of_base.calc_signal_filt_td(
                self._channel_name,
                template_tag=self._template_tag
            )

        # calculate Q vector
        self._calc_q_vector()

        # calculate amp and chi2
        self._calc_amps()
        self._calc_chi2(polarity_constraint=polarity_constraint)


        
    def get_fit(self):
        """
        FIXME
        #docstrings need to be added with dimensions
        #returns that send back the ofamp chi2 and t need to be added
        #interpolate option needs to be added

        FIXME need to add interpolate option
        
        """

        # min chi2 index
        min_index = np.argmin(self._chi2_allt)
        
        self._of_amp = self._amps_allt[min_index]
        self._of_t0 =  (self._time_combinations[min_index, 1]/self._fs
                        - self._time_combinations[min_index, 0]/self._fs)
        self._of_chi2 = self._chi2_allt[min_index]
        self._index_first_pulse = self._time_combinations[min_index, 0]
        self._index_second_pulse =  self._time_combinations[min_index, 1]

        nbins = self._of_base.nb_samples()
        self._of_chi2_per_DOF = self._of_chi2/(self._nchans*nbins)

        return self._of_amp, self._of_t0, self._of_chi2

    def _calc_q_vector(self):
        """
        Calculate Q vector
        """

        
        signal_filt_td = (
            self._of_base.signal_filt_td(
                self._channel_name,
                template_tag=self._template_tag)
        )
            
        # initialize
        self._q_vector = np.zeros(
            (self._ntmps, self._time_combinations[:,0].shape[0] )
        )
        
        for itmps in range(self._ntmps):
            self._q_vector[itmps,:] = (
                signal_filt_td[itmps][
                    self._time_combinations[:,self._template_group_ids[itmps]]
                ]
            )
        
    def _calc_amps(self):
        """
        Calc amplitude all times
       
        """

        if self._q_vector is None:
            self._calc_q_vector()
            
        
        self._amps_allt = np.zeros((self._time_combinations[:,0].shape[0],
                               self._ntmps))

        for itmps in range(self._ntmps):
            for jtmps in range(self._ntmps):
                self._amps_allt[:,itmps] += (self._p_matrix_inv[:,itmps,jtmps]
                                        * self._q_vector[jtmps,:])
                


    def _calc_chi2(self, polarity_constraint=False ):
        '''
        FIXME
        docstrings and dimensions need to be added
        dim: [ntmps, nbins]
        '''

        if self._amps_allt is None:
            self._calc_amps()

        
        # get signal matrix fft
        signal_fft = self._of_base.signal_fft(self._channel_name)

        # get invert covariant matrix
        icovf = self._of_base.icovf(self._channel_name)


        # chi2 base (time independent scalar on the sum over
        # all channels & freq bins)
        chi2base = np.einsum('kf,kjf,jf->', 
                             signal_fft.conjugate(), 
                             icovf, 
                             signal_fft, 
                             optimize=True)
     
        chi2base = np.real(chi2base)

        #chi2_t is the time dependent part
        chi2_t = np.zeros_like(self._time_combinations[:,0])
        chi2_when_one_deviates =  np.zeros_like(self._time_combinations[:,0])

        chi2_t = np.real(np.sum(
            np.conjugate(self._q_vector) * self._amps_allt.T,
            axis=0)
        )

        if polarity_constraint: # this is zero when polarity constrain is not used
            
            chi2_polarity = np.zeros(self._amps_allt.shape[0])
            
            for ibins in range(chi2_polarity.shape[0]):
                chi2_polarity[ibins] = (self._amps_allt.T[:,ibins]
                                        @self._p_matrix[ibins,:,:]
                                        @self._amps_allt.T[:,ibins])

            chi2_when_one_deviates = chi2_polarity - (
                np.real(np.sum(
                    np.conjugate(self._q_vector) * self._amps_allt.T,
                    axis=0)
                )
            )
                
        self._chi2_allt = chi2base - chi2_t - chi2_when_one_deviates


    



import os
import numpy as np
from qetpy import IBIS
import pytest


THIS_DIR = os.path.dirname(os.path.abspath(__file__))

def test_ibis():
    pathtodata = os.path.join(THIS_DIR, "data/test_iv_data.npz")
    testdata = np.load(pathtodata)
    testdata = np.load(pathtodata)
    dites = testdata["dites"][0:1]
    dites_err = testdata["dites_err"][0:1]
    rshunt = 5e-3
    ib = testdata["vb"][0:1]/rshunt
    ib_err = testdata["vb_err"][0:1]*0
    
    norminds = range(0,4)
    scinds = range(12,15)

    dites_bad = testdata["dites"][2:3]
    dites_err_bad = testdata["dites_err"][2:3]
    ib_bad = testdata["vb"][2:3]/rshunt
    ib_err_bad = testdata["vb_err"][2:3]*0

    
    
    ivobj_bad = IBIS(dites=dites_bad, dites_err=dites_err_bad,ibias=ib_bad, 
                        ibias_err=ib_err_bad, rsh=5e-3, rsh_err=5e-4,
                        rp_guess=5e-3, rp_err_guess=0, 
                        chan_names=['a','b','c'], fitsc=True,
                        normalinds=norminds, scinds=scinds)
    with pytest.raises(ValueError):
        ivobj_bad.analyze()

    ivobj = IBIS(dites=dites, dites_err=dites_err,ibias=ib, 
                        ibias_err=ib_err, rsh=5e-3, rsh_err=5e-4,
                        rp_guess=5e-3, rp_err_guess=0, 
                        chan_names=['a','b','c'], fitsc=True,
                        normalinds=norminds, scinds=scinds)
    ivobj.analyze()
    ivobj.plot_all_curves()
    
    test_val1 = np.all([np.all(np.isclose(ivobj.rp,np.array([[0.00703892, 0.00578437, 0.00722202]]))),
    np.all(np.isclose(ivobj.rnorm,np.array([[0.32464424, 0.32076859, 0.32058897]]))),
    np.all(np.isclose(ivobj.rp_err,np.array([[0.00070389, 0.00057844, 0.0007222 ]]))),
    np.all(np.isclose(ivobj.rnorm_err,np.array([[0.03367567, 0.03316034, 0.03328893]])))])
    
    ivobj = IBIS(dites=dites, dites_err=dites_err,ibias=ib, 
                        ibias_err=ib_err, rsh=5e-3, rsh_err=5e-4,
                        rp_guess=np.array([[0.00703892, 0.00578437, 0.00722202]]), 
                        rp_err_guess=np.array([[0.00070389, 0.00057844, 0.0007222 ]]), 
                        chan_names=['a','b','c'], fitsc=False,
                        normalinds=norminds, scinds=scinds)
    ivobj.analyze()
    ivobj.plot_all_curves()
    
    test_val2 = np.all([np.all(np.isclose(ivobj.rp,np.array([[0.00703892, 0.00578437, 0.00722202]]))),
    np.all(np.isclose(ivobj.rnorm,np.array([[0.32464424, 0.32076859, 0.32058897]]))),
    np.all(np.isclose(ivobj.rp_err,np.array([[0.00070389, 0.00057844, 0.0007222 ]]))),
    np.all(np.isclose(ivobj.rnorm_err,np.array([[0.03367567, 0.03316034, 0.03328893]])))])
    
    assert np.all([test_val1, test_val2])



import numpy as np
from scipy.optimize import least_squares
import matplotlib.pyplot as plt
from qetpy.utils import shift, interpolate_of, argmin_chisq
from qetpy.core import OFBase
from numpy.linalg import pinv as pinv
from qetpy.utils import convert_channel_name_to_list, convert_channel_list_to_name


__all__ = ['OFnxm',
           'get_time_offset_nxm']

class OFnxm:
    """
    N channels / M templates Optimal Fitler 
    (single OF delay)
    
    """
    def __init__(self, of_base=None,  channels=None,
                 templates=None, template_tag=None,
                 csd=None, sample_rate=None,
                 pretrigger_msec=None, pretrigger_samples=None,
                 integralnorm=False,
                 verbose=True):

        """
        Initialize OFnxm

        Parameters
        ----------

        of_base : OFBase object, optional
           OF base with pre-calculations
           Default: instantiate base class within OFnxm

        channels : str or list
          channels as ordered list or "|" separated string
          such as "channel1|channel2"

        templates : 3D ndarray [nchan, ntmp, nbins] (optional)
          multi-template array used for OF calculation, can be
          None if already in of_base, otherwise required
          

        template_tag : str (optional)
            Tag associated with templates (to store in of_base)
       
        csd : ndarray, optional
          psd array used for OF calculation, can be
          None if already in of_base, otherwise required


        sample_rate : float, optional if of_base is not None
          The sample rate of the data being taken (in Hz)
          Only Required if of_base=None

        pretrigger_samples : int, optional if of_base is not None
            Number of pretrigger samples
            Default: use pretrigger_msec or if also None,
                     use 1/2 trace length

        pretrigger_msec : float, optional if of_base is not None
            Pretrigger length in ms (if  pretrigger_samples is None)
            Default: 1/2 trace length

        integralnorm : bool, optional
            If set to True, then  template will be normalized
            to an integral of 1, and any optimum filters will
            instead return the optimum integral in units of Coulombs.
            If set to False, then the usual optimum filter amplitudes
            will be returned (in units of Amps).
            Default=False

        verbose : bool, optional
            Display information
            Default=True

        Return
        ------

        """
        # verbose
        self._verbose = verbose
        
        # channels
        if channels is None:
            raise ValueError('ERROR: "channels" argument required!')
        
        self._channel_name = convert_channel_list_to_name(channels)
        self._channel_list = convert_channel_name_to_list(channels)
        self._nchans = len(self._channel_list)
             
        # initialize number of templates (per channel)
        self._ntmps = None
        self._template_tag = None

        # template tag
        self._template_tag = None
        if template_tag is not None:
            self._template_tag = template_tag
        else:
            if templates is None:
                raise ValueError('ERROR: template tag required if '
                                 'templates array not provided!')
            # assign "default"
            self._template_tag = 'default'
            
                            
        # Instantiate OF base (if None) and add templates
        self._of_base = of_base

        # instantiate OF base if None
        if of_base is None:

            # templates required
            if templates is None:
                raise ValueError('ERROR:  Either "of_base" or '
                                 '"templates" argument required!')

            
            # check required parameters
            if sample_rate is None:
                raise ValueError('ERROR in OFnxm: sample rate required!')

            if (pretrigger_msec is None
                and pretrigger_samples is None):
                raise ValueError('ERROR in OFnxm: '
                                 'pretrigger (msec or samples) required!')
            # instantiate
            self._of_base = OFBase(sample_rate, verbose=verbose)

        else:

            if (templates is None and
                self._of_base.template(self._channel_name,
                                       self._template_tag) is None):
                raise ValueError(
                    f'ERROR: "templates" for channel {self._channel_name} '
                    f'and tag "{self._template_tag}" not found '
                    f'in of_base object !')
            
        # add template to base object if templates not None
        calc_phi = False
        if templates is not None:

            # check array
            if not isinstance(templates, np.ndarray):
                raise ValueError('ERROR: Expecting "templates" to be '
                                 'a numpy array')

            if templates.ndim == 1:
                templates = templates[np.newaxis, np.newaxis, :]  # 1 channel, 1 template
            elif templates.ndim == 2:
                templates = templates[np.newaxis, :, :]            # 1 channel, multiple templates
            elif templates.ndim != 3:
                raise ValueError('ERROR: "templates" input must be 1D, 2D, or 3D')    

            if templates.shape[0] != self._nchans:
                raise ValueError(f'ERROR: Expecting "templates" to have '
                                 f'shape[0]={self._nchans}!')
            # calc phi
            calc_phi = True
            
            fs = self._of_base.sample_rate
            if pretrigger_msec is not None:
                pretrigger_samples = int(round(fs*pretrigger_msec/1000))
                                    
            # add to OF 
            self._of_base.add_template(
                self._channel_name, templates, self._template_tag,
                pretrigger_samples=pretrigger_samples,
                integralnorm=integralnorm,
                overwrite=True
            )
                
        else:

            templates =  self._of_base.template(
                self._channel_name,
                template_tag=self._template_tag
            )
            
            if templates is None:
                raise ValueError(
                    f'ERROR: No template with tag '
                    f'"{template_tag}" found for channel '
                    f'{self._channel_name} in OF base!'
                )


        # store as internal variable 
        self._templates =  self._of_base.template(
            self._channel_name,
            template_tag=self._template_tag
        )
        
        self._ntmps = self._templates.shape[1]
        if (self._templates.shape[0] != self._nchans):
            raise ValueError(f'ERROR: Inconsistent number of channels: '
                             f'templates: {self._templates.shape[0]}, '
                             f'input channels {self._channel_name}: '
                             f'{self._nchans}')
        

        
        # add noise to base object
        if csd is not None:

            if self._verbose:
                print('INFO: Adding noise CSD to OF base object')
                
            self._of_base.set_csd(self._channel_name, csd=csd)
            
        elif self._of_base.csd(self._channel_name) is None:
            raise ValueError('ERROR: No csd found in OF base object.'
                             ' Add csd argument!')
                  
        # calculate optimal filter and weight matrix (if not done)
        if (calc_phi
            or self._of_base.phi(self._channel_name,
                                 self._template_tag) is None):
            if self._verbose:
                print('INFO: Calculating optimal filter!')
                
            self._of_base.calc_phi(self._channel_name,
                                   self._template_tag)
            
        # initialize fit results
        #variables need to be added (chi2, amp, ntmp,nchan,pretriggersamples, etc.)
        self._nbins = self._of_base._nbins
        self._pretrigger_samples = self._of_base.nb_pretrigger_samples(
            self._channel_name, self._template_tag
        )

        self._fs = self._of_base.sample_rate

        self._amps_alltimes_rolled = None
        self._amps_alltimes = None
        self._chi2_alltimes = None
        self._chi2_alltimes_rolled = None

        self._of_amp_withdelay = None
        self._of_chi2_withdelay = None
        self._of_t0_withdelay = None

        
    def clear(self):
        """
        clear signal data
        """

        self._amps_alltimes_rolled = None
        self._amps_alltimes = None
        self._chi2_alltimes = None
        self._chi2_alltimes_rolled = None
        
        self._of_amp_withdelay = None
        self._of_chi2_withdelay = None
        self._of_t0_withdelay = None
        
        

    def calc(self, signal=None):
        """
        OF NxM calculations
  
        Parameters
        ----------

        signal : nd array (optional)
           2D array [nchan, nsamples]
           NEEDS TO FOLLOW ORDER of channels argument
           used in the OFnxm instantiation
            
            optional if already in OF base
           
        Return:
        -------
        None
 
        """
                
        # clear internal (signal) data
        self.clear()

                
        # update signal and do preliminary (signal) calculations
        # (like fft signal, filtered signal, signal matrix ...
        if signal is not None:

            # check array
            if not isinstance(signal, np.ndarray):
                  raise ValueError('ERROR: signal should be a '
                                   'numpy array!')

            
            # check dimension
            if signal.ndim == 1:
                signal = signal[np.newaxis, :]
            
            # check if correct size
            if signal.shape[0] != self._nchans:
                raise ValueError('ERROR: Wrong number of channels '
                                 'in "signal" array!')
            
            self._of_base.clear_signal()

            # update signal
            self._of_base.update_signal(
                self._channel_name,
                signal,
                calc_fft=True)

        # calculate signal filter
        if self._of_base.signal_filt(
            self._channel_name,
            template_tag=self._template_tag
        ) is None:
            
            self._of_base.calc_signal_filt(
                self._channel_name,
                template_tag=self._template_tag
            )
            
            # calculate signal filter td
            self._of_base.calc_signal_filt_td(
                self._channel_name,
                template_tag=self._template_tag
            )
            
              
        # calculate amplitude all times
        self._calc_amp_allt()

        # calculate chi2 all times
        self._calc_chi2_allt()


        
    def get_fit_nodelay(self):
        """
        New nodelay version of NxM fits. Just returns the nxm best fit
        at the index of pretrigger_samples
        """

        amp_all = self._amps_alltimes_rolled
        chi2_all = self._chi2_alltimes_rolled
        pretrigger_samples = self._pretrigger_samples
        
        amp = amp_all[:,pretrigger_samples]
        t0 = (pretrigger_samples)/self._fs
        chi2 = chi2_all[pretrigger_samples]

        # save
        self._of_amp_nodelay = amp
        self._of_chi2_nodelay = chi2
        self._of_t0_nodelay = t0

        return amp, t0, chi2
    
    def get_fit_withdelay(self, 
                          window_min_from_trig_usec=None,
                          window_max_from_trig_usec=None,
                          window_min_index=None,
                          window_max_index=None,
                          lgc_outside_window=False,
                          pulse_direction_constraint=0,
                          interpolate_t0=False):
        """
        FIXME
        #docstrings need to be added with dimensions
        #returns that send back the ofamp chi2 and t need to be added
        #interpolate option needs to be added
        """

        amp_all = self._amps_alltimes_rolled
        chi2_all = self._chi2_alltimes_rolled
        pretrigger_samples = self._pretrigger_samples
        
        # mask pulse direction
        constraint_mask = None
        if (pulse_direction_constraint==1 or pulse_direction_constraint==-1):
            constraint_mask=(amp_all*pulse_direction_constraint>0)

        # find index minimum chisq within window
        window_min = None
        if window_min_from_trig_usec is not None:
            window_min = np.floor(pretrigger_samples
                                + window_min_from_trig_usec*self._fs*1e-6)
        elif window_min_index is not None:
            window_min = window_min_index

        if window_min is not None and window_min<0:
            window_min = 0
        window_max = None
        if window_max_from_trig_usec is not None:
            window_max = np.ceil(pretrigger_samples
                                + window_max_from_trig_usec*self._fs*1e-6)
        elif window_max_index is not None:
            window_max = window_max_index

        if window_max is not None and window_max>self._nbins:
            window_max = self._nbins

        
        if  window_min is not None:
             window_min = int(window_min)
             
        if  window_max is not None:
             window_max = int(window_max)

            
        #argmin_chisq will minimize along the last axis
        #chi2_all dim [ntmp,nbins]
           
        bestind = argmin_chisq(
            chi2_all,
            window_min=window_min,
            window_max=window_max,
            lgc_outside_window=lgc_outside_window,
            constraint_mask=constraint_mask)
        
        #need to add interpolate option
        amp = amp_all[:,bestind]
        t0 = (bestind-pretrigger_samples)/self._fs
        chi2 = chi2_all[bestind]

        # save
        self._of_amp_withdelay = amp
        self._of_chi2_withdelay = chi2
        self._of_t0_withdelay = t0

        return amp, t0, chi2


    def get_fit_overlay(self, amp, t0):
        """
        Get fit overlay to display fit
        """

        # get template
        templates = self._of_base.template(self._channel_name,
                                           self._template_tag)
        
        if templates is None:
            return []

        # sample rate 
        fs = self._of_base.sample_rate
        
        
        # fit overlay
        shift = int(t0 * fs)
        rolled_templates = np.roll(templates, shift, axis=-1)
        fit_overlay = np.sum(rolled_templates * amp[None, :, None], axis=1)


        return fit_overlay

    

    def _calc_amp_allt(self):
        """
        FIXME
        #docstrings need to be added with dimensions
        dim: [ntmps, nbins]
        """
  
        # calc_signal_filt_mat_td checks that
        # signal_filts_mat is calculated first

        if self._of_base.signal_filt_td(
                self._channel_name,
                template_tag=self._template_tag
        ) is None:
            if self._verbose:
                print('INFO: filtered signal not available. '
                      'Calculating it')
                
            self._of_base.calc_signal_filt_td(
                self._channel_name,
                template_tag=self._template_tag)

        # signal filt 
        signal_filt = self._of_base.signal_filt_td(
            self._channel_name,
            template_tag=self._template_tag
        )
            
        # inverted weight matrix
        iw = self._of_base.iweight(self._channel_name,
                                   self._template_tag)

        # calculate
        self._amps_alltimes = (iw @ signal_filt)
           
        # roll with pretrigger_samples 
        temp_amp_roll = np.zeros_like(self._amps_alltimes)
        for itmp in range(self._ntmps):
            temp_amp_roll[itmp,:] = np.roll(self._amps_alltimes[itmp,:],
                                            self._pretrigger_samples,
                                            axis=-1)
        self._amps_alltimes_rolled = temp_amp_roll


        
    def _calc_chi2_allt(self):
        """
        FIXME
        docstrings and dimensions need to be added
        dim: [ntmps, nbins]
        """

        # get signal fft 
        signal_fft = self._of_base.signal_fft(self._channel_name)
        
        # inverted cov matrix
        icov_f = self._of_base.icovf(self._channel_name)

        # amplitude all time
        temp_amp_allt = self._amps_alltimes.copy()

        # signal filt
        filt_signal_t = self._of_base.signal_filt_td(
            self._channel_name,
            template_tag=self._template_tag
        )

        # calculate chi2

        #chi2base is a time independent scalar on the sum over all channels & freq bins
        chi2base = np.einsum('kf,kjf,jf->', signal_fft.conjugate(), icov_f,
                             signal_fft, optimize=True)

        chi2base = np.real(chi2base)
        chi2_t = np.zeros_like(temp_amp_allt)
        chi2_t = np.real(np.sum(temp_amp_allt*filt_signal_t, axis=0))
        
        #this sums along the template
        #dim [ntmp, nbins]
        #chi2_t is the time dependent part
        self._chi2_alltimes = chi2base - chi2_t
        self._chi2_alltimes_rolled = np.roll(self._chi2_alltimes,
                                             self._pretrigger_samples,
                                             axis=-1)
                                             
                                             
def get_time_offset_nxm(csd, template_1, template_2, fs=1.25e6, start_time=10e-3):
    """
    Calculates the offset between two different NxM templates, so that
    different templates will trigger at nominally the same time. There
    will still be some offset due to differences between the real
    event and the template, but the average difference should be zero.
    
    Parameters
    ----------
    
    csd : 2x2xn numpy array
        Array of size 2x2x(number of samples per trace) input into the 
        2x1 optimal filter.
        
    template_1 : 2x1xn numpy array
        Array with size 2x1x(number of samples per trace) input into the 
        2x1 optimal filter. This should be the ''main trigger,'' i.e. the
        t0 output by this function will tell you how far to offset
        template_2 to make it consistent with template_1.
        
    template_2 : 2x1xn numpy array
        Array with size 2x1x(number of samples per trace) input into the 
        2x1 optimal filter. This should be the ''secondary trigger,'' i.e.
        the t0 output by this function will tell you how far to offset
        template_2 to make it consistent with template_1.
        
    fs : float, optional
        Sampling frequency (e.g. defaults to 1.25 MHz)
        
    start_time : float, optional
        The start time of the event, defaults to 10 ms.
        
    Return:
    -------
    
    t0 : float
        The time offset needed to be added to the start time of
        template_2 to make it consistent with template_1
    """

    channels= ['channel1', 'channel2']
    ofnxm_1 = OFnxm(channels=channels, templates=template_1,
                    csd=csd, sample_rate=fs, 
                    pretrigger_samples=int(start_time*fs), verbose=False)
    
    ofnxm_1.calc(signal=template_2[:,0,:])
    amp, t0, chi2 = ofnxm_1.get_fit_withdelay()
    return -t0




import numpy as np
from math import ceil, floor
from qetpy.utils import shift, interpolate_of, argmin_chisq
from qetpy.utils import fft, ifft, fftfreq, rfftfreq
from qetpy.utils import convert_channel_name_to_list, convert_channel_list_to_name
from numpy.linalg import pinv as pinv
import time
import copy


__all__ = ['OFBase']

class OFBase:
    """
    Multiple channels - multiple templates optimal filter base class.
    Calculate FFT, optimal filter, and filtered traces.

    Each template has a name tag.  The number of samples should be the same
    between templates, psd, and traces and accross channels for all algorithms

    """
    def __init__(self, sample_rate,
                 verbose=True):
        """
        Initialization of the optimum filter base class

        Parameters
        ----------

        sample_rate : float
            The sample rate of the data being taken (in Hz).

        verbose : bool, optional
            Display information
            Default=True


        Return
        ------
        None

        """
        self._debug = False
        self._verbose = verbose
        self._fs = sample_rate

        # initialize frequency spacing of FFT and frequencies
        self._df = None
        self._fft_freqs = None

        # number of samples (needs to be same for all algorithms)
        # instantiate a new of base if different trace lenght!
        self._nbins = None

        # initialize templates dictionaries
        # (time domain and FFT)
        self._templates = dict()
        self._templates_fft = dict()

        # time constraint for NxMx2
        self._time_constraints = dict()
    
        # initialize two-sided noise csd (in Amps^2/Hz)
        self._csd = dict()

        # initialize the inverted covariance matrix (csd)
        self._icovf = dict()
        
        # pretrigger length (can be different between
        # templates)
        self._pretrigger_samples = dict()

        # initialize calculated optimal filter and norm
        # (independent of signal)
        self._phis = dict()
        self._weights = dict()
        self._iweights = dict()
        self._p_matrix  = dict()
        self._p_matrix_inv  = dict()
      
        # initialize signal
        self._signals = dict()
        self._signals_fft = dict()
     
        # initialize (optimal) filtered  signals and templates
        # (frequency domain and  converted back to time domain)
        # dict key = template tag

        # filtered signal
        self._signals_filts = dict()
        self._signals_filts_td = dict()

        # filtered template
        self._template_filts = dict()
        self._template_filts_td = dict()
    

    @property
    def verbose(self):
        return self._verbose

    @property
    def sample_rate(self):
        return self._fs


    # =====================================================
    # Get data
    # =====================================================
    def df(self):
        """
        Frequency resolution
        """

        return self._df
    
    def nb_samples(self):
        """
        Number of samples
        (same all channels)
        """

        return self._nbins
    
    def fft_freqs(self):
        """
        Return frequency array
        """
        return self._fft_freqs

    def nb_pretrigger_samples(self, channels, template_tag='default'):
        """
        get pretrigger samples
        """

        # convert to string if needed (if str, return same string)
        channel_name = convert_channel_list_to_name(channels)
        
        pretrigger_samples = None
        if (channel_name in self._pretrigger_samples
            and template_tag in self._pretrigger_samples[channel_name]):
            pretrigger_samples = (
                self._pretrigger_samples[channel_name][template_tag]
            )
                
        return pretrigger_samples

    
    def channels_with_noise_spectrum(self):
        """
        Get channel list based on 
        PSD and CSD
        """
        channels = list(self._csd.keys())
        return channels

    
    def template_tags(self, channels):
        """
        get template tags
        
        Parameters
        ----------
        channels : str or array-like
            if multiple channels:
                list or  "|" separated string
                such as "channel1|channel2"

        Return
        ------
        tag : list
         list with template tags

        """

        # convert to string if needed (if str, return same string)
        channel_name = convert_channel_list_to_name(channels)
  
        tags = []
        if channel_name in self._templates.keys():
            tags = list(self._templates[channel_name].keys())
        
        return tags

        
        
    def template(self, channels,
                 template_tag='default',
                 squeeze_array=False):
        """
        Get template (s) in time domain for the specified tag

        Parameters
        ----------

        channels : str or array-like
            if multiple channels:
                list or  "|" separated string
                such as "channel1|channel2"

        template_tag : str (single channel)
                        or 2D array (multiple channels)
          template tags/id

        squeeze_array : boolean
          array store as 3D, flag to sequeeze to 1D or 2D
                   
        Return
        ------

        template(s) : ndarray 1D, 2D, or 3D 
         template trace(s)  in time domain
         if multiple channels:
            template dim: [nchans, ntmps, nbins]

        """

        # channel name and list
        channel_name = convert_channel_list_to_name(channels)

        # check available
        if (channel_name not in self._templates
            or template_tag not in self._templates[channel_name]):
            return None
                 
        # get template matrix (3D)
        template = self._templates[channel_name][template_tag]
        if squeeze_array:
            template = np.squeeze(template)

        return template
    
                        
 
    def template_fft(self, channels,
                     template_tag='default',
                     squeeze_array=False):
        """
        Get template FFT for the specified tag(s)

        Parameters
        ----------
        
        channels : str or array-like
            if multiple channels:
                list or  "|" separated string
                such as "channel1|channel2"

        template_tag : str (single channel)
                        or 2D array (multiple channels)
          template tags/id


        Return
        ------

        template(s) : ndarray
         template FFTs
         if multiple channels:
            template dim: [nchans, ntmps, nbins]

        """

        # channel name
        channel_name = convert_channel_list_to_name(channels)

        # check available
        if (channel_name not in self._templates_fft
            or template_tag not in self._templates_fft[channel_name]):
            return None
                   
        # get template matrix (3D)
        template_fft = self._templates_fft[channel_name][template_tag]
        if squeeze_array:
            template_fft = np.squeeze(template_fft)

        return template_fft
        
    def template_group_ids(self, channels,
                           time_constraints_tag='default'):
        """
        Get NxMx2 template group ids for specified channel
        """
        
        # channel name
        channel_name = convert_channel_list_to_name(channels)

        # check available
        if (channel_name not in self._time_constraints
            or time_constraints_tag not in self._time_constraints[channel_name]):
            raise ValueError(f'ERROR: NxMx2 time constraints not available for '
                             f'channel {channel_name} and tag '
                             f'{time_constraints_tag}')
        
        # get constraints
        time_constraints = self._time_constraints[channel_name][time_constraints_tag]

        if 'template_group_ids' not in time_constraints:
            raise ValueError(f'ERROR: "template_group_ids" not available for '
                             f'channel {channel_name} and tag '
                             f'{time_constraints_tag}')

        return time_constraints['template_group_ids']


    def time_combinations(self, channels,
                          time_constraints_tag='default'):
        """
        Get NxMx2 time_combinations for specified channel
        """
        
        # channel name
        channel_name = convert_channel_list_to_name(channels)

        # check available
        if (channel_name not in self._time_constraints
            or time_constraints_tag not in self._time_constraints[channel_name]):
            return None
        
        # get constraints
        time_constraints = self._time_constraints[channel_name][time_constraints_tag]

        if 'time_combinations' not in time_constraints:
            return None

        return time_constraints['time_combinations']
    
        
    def psd(self, channel):
        """
        Get psd

        Parameters
        ----------
        channel : str
          channel name


        Return
        ------

        psd : ndarray
         noise PSD

        """

        channel_name = convert_channel_list_to_name(channel)
        
        if channel_name not in self._csd.keys():
            return None
        
        array = self._csd[channel_name]
        array = np.squeeze(array)
        
        return array

    def csd(self, channels):
        """
        Get csd

        Parameters
        ----------

        channels : str or array-like
                list or  "|" separated string
                such as "channel1|channel2"
        

        Return
        ------

        csd: ndarray
         noise CSD

        """

        # convert to string if needed (if str, return same string)
        channel_name = convert_channel_list_to_name(channels)

        # get add and return
        if channel_name in self._csd.keys():
            csd = self._csd[channel_name]
            return csd.copy()
        else:
            return None


    def icovf(self, channels):
        """
        Get the inverted noise covariance (csd) matrix between channels. 
        
        Parameters
        ----------
        
        channels : str or array-like
                list or  "|" separated string
                such as "channel1|channel2"
                
        Return
        ------
        
        icovf: ndarray, dimn: [nchans,nchans,nbins]
            The inverted csd in units 1/A^2. 
        """
        
        # convert to string if needed (if str, return same string)
        channel_name = convert_channel_list_to_name(channels)
        
        # return icovf
        if channel_name in self._icovf.keys():
            return self._icovf[channels]
        else:
            return None

        
    def is_signal_stored(self, channels):
        """
        Check if signals have been stored for a specific
        channel (or multiple channels)
         
        Parameters
        ----------
        channel : str or array-like
             single channel: channel name              
             multi channels: array-like or  "|" separated string
                such as "channel1|channel2"

        Return
        ------
        is_stored : boolean
            if all signals stored return True
        """

        if self._nbins is None:
            return False
        
        channel_list = convert_channel_name_to_list(channels)

        for ichan, chan in enumerate(channel_list):
            if chan not in self._signals:
                return False

        return True
        
    def signal(self, channels, squeeze_array=False):
        """
        Get signal trace(s) in time domain
      
        Parameters
        ----------
        channels : str or array-like
             single channel: channel name                
             multi channels: array-like or  "|" separated string
                such as "channel1|channel2"
       '
        Return
        ------
        signal : ndarray 1D or 2D
          time domain signal trace
          if multiple channels or squeeze_array=False
             ndarray = [nchans, nsamples]
        """

        # check nsamples
        if self._nbins is None:
            raise ValueError('ERROR: No signal stored!')

        
        # convert to string if needed (if str, return same string)
        channel_name = convert_channel_list_to_name(channels)
        channel_list = convert_channel_name_to_list(channels)
        nchans = len(channel_list)

        signal_array = np.zeros((nchans, self._nbins),
                                dtype='float64')
        
        for ichan, chan in enumerate(channel_list):
            if chan not in self._signals:
                raise ValueError(f'ERROR: No signal stored for '
                                 f'{chan} !')
            signal_array[ichan,:] = self._signals[chan]

        if squeeze_array:
             signal_array  = np.squeeze(signal_array)

        return signal_array
                
        
    def signal_fft(self, channels, squeeze_array=False):
        """
        Get current signal(s) trace FFT

        Parameters
        ----------

        channels : str or array-like
             single channel: channel name                
             multi channels: array-like or  "|" separated string
                such as "channel1|channel2"

        Return
        ------
        signals_fft : ndarray
          signal trace FFT
          if multiple channels or squeeze_array=False
             ndarray = [nchans, nsamples]

        """

        # check nsamples
        if self._nbins is None:
            raise ValueError('ERROR: No signal  stored!')

        
        # convert to string if needed (if str, return same string)
        channel_name = convert_channel_list_to_name(channels)
        channel_list = convert_channel_name_to_list(channels)
        nchans = len(channel_list)
        
        signal_array = np.zeros((nchans, self._nbins),
                                dtype=np.complex128)
        
        for ichan, chan in enumerate(channel_list):
            if chan not in self._signals_fft:
                raise ValueError(f'ERROR: No signal FFT stored for '
                                 f'{chan} !')
            signal_array[ichan,:] = self._signals_fft[chan]

        if squeeze_array:
             signal_array  = np.squeeze(signal_array)

        return signal_array

    
    def phi(self, channels, template_tag='default',
            squeeze_array=False):
        """
        Get 'optimal filter' (phi) for a specified
        template tag, depends only
        on template and PSD

        Parameters
        ----------
        channel : str
          channel name

        template_tag : str or 2D array (optional)
          template tag/id
          default: 'default'

        Return
        ------

        phi : ndarray
          'optimal filter' values

        """

        # channel name and list
        channel_name = convert_channel_list_to_name(channels)


        # check availability
        if (channel_name not in self._phis
            or template_tag not in self._phis[channel_name]):
            return None
   
        # get array
        phi = self._phis[channel_name][template_tag]

        # squeeze
        if squeeze_array:
            phi = np.squeeze(phi)

        return phi

  
    def weight(self, channels, template_tag='default',
               squeeze_array=False):
        """
        Get the weighting matrix for the specified combination
        of channels. 
        
        Parameters
        ----------
        
        channels : str or array-like
            channels as "|" separated string
            such as "channel1|channel2"

        template_tag : str
             tag associate with templates

        Returns
        -------
        
        w : ndarray, dimn: [ntmps,ntmps]
            The inverted matrix which describes how we weight each
            template w/ respect to one another. This is "norms" in the 
            1x1. 
        """
        
        # channel name and list
        channel_name = convert_channel_list_to_name(channels)

        # check if available
        if (channel_name not in self._weights
            or template_tag not in self._weights[channel_name]):
            return None
            
        w = self._weights[channel_name][template_tag]

        if squeeze_array:
            w = np.squeeze(w)
            if (w.ndim==1 and len(w)==1):
                w = float(w[0])
            
        return w
    
    def iweight(self, channels, template_tag='default',
                squeeze_array=False):
        """
        Get the inverted weighting matrix for the specified combination
        of channels. 
        
        Parameters
        ----------
        
        channels : str or array-like
            channels as "|" separated string
            such as "channel1|channel2"

        template_tag : str
             tag associate with templates

        Returns
        -------
        
        iw : ndarray, dimn: [ntmps,ntmps]
            The inverted matrix which describes how we weight each
            template w/ respect to one another. This is "norms" in the 
            1x1. 
        """
        
        # channel name and list
        channel_name = convert_channel_list_to_name(channels)

        # check if available
        if (channel_name not in self._iweights
            or template_tag not in self._iweights[channel_name]):
            return None
            
        iw = self._iweights[channel_name][template_tag]

        if squeeze_array:
            iw = np.squeeze(iw)
            if (iw.ndim==1 and len(iw)==1):
                iw = float(iw[0])
        
        return iw
    

    def p_matrix_inv(self, channels,
                     template_tag='default',
                     time_constraints_tag='default'):
        """
        Get NxMx2 p_matrix INVERTED
        
        dim: [ntmps, ntmps]
        """

        # convert to nname
        channel_name = convert_channel_list_to_name(channels)

        # combined tag
        combined_tag = f'{template_tag}_{time_constraints_tag}'


        # check
        if (channel_name not in self._p_matrix_inv
            or combined_tag not in self._p_matrix_inv[channel_name]):
            return None

        # return
    
        return self._p_matrix_inv[channel_name][combined_tag]
    

    
    def p_matrix(self, channels,
                 template_tag='default',
                 time_constraints_tag='default'):
        """
        Get NxMx2 p_matrix
        
        dim: [ntmps, ntmps]
        """

        # convert to nname
        channel_name = convert_channel_list_to_name(channels)

        # combined tag
        combined_tag = f'{template_tag}_{time_constraints_tag}'

        # check
        if (channel_name not in self._p_matrix
            or combined_tag not in self._p_matrix[channel_name]):
            return None
        
        # return
        return self._p_matrix[channel_name][combined_tag]


    
 
    def signal_filt(self, channels,
                    template_tag='default',
                    squeeze_array=False):
        """
        Get (optimal) filtered signal in frequency domain
        for the specified template tag

        signal_filt = phi*signal_fft

        Parameters
        ----------
        channel : str or list
          channel name

        template_tag : str
          tag associated with template
          default: 'default'


        Returns
        -------
        signal_filt  : ndarray
           optimal filtered signal

        """

        # channel name and list
        channel_name = convert_channel_list_to_name(channels)


        # check availability
        if (channel_name not in self._signals_filts
            or  template_tag not in self._signals_filts[channel_name]):
            return None
            
        # get array and squeeze
        array = self._signals_filts[channel_name][template_tag]
        if squeeze_array:
            array = np.squeeze(array)
        

        return array

    def signal_filt_td(self, channels,
                       template_tag='default',
                       squeeze_array=False):
        """
        Get (optimal) filtered signal converted back to time domain
        for the specified template tag

        Parameters
        ----------
        channels : str or list of str
          channel name (single channel or NxM)

        template_tag : str, optional
          template tag/id
          default: 'default'

        Returns
        -------
        signal_filt_td  :ndarray
           optimal filtered signal in time domain

        """

        # channel name 
        channel_name = convert_channel_list_to_name(channels)

        
        # check availability
        if (channel_name not in self._signals_filts_td
            or  template_tag not in self._signals_filts_td[channel_name]):
            return None
                    
        # get array and squeeze
        array = self._signals_filts_td[channel_name][template_tag]
        if squeeze_array:
            array = np.squeeze(array)

        return array
    


    def template_filt(self, channels,
                      template_tag='default',
                      squeeze_array=False):
        """
        (optimal) filtered template 

        Get (optimal) filtered template in frequency domain
        for a specified template tag

        Parameters
        ----------
        channels : str
          channel name

        template_tag : str, optional
          template tag/id
          default: 'default'

        Returns
        -------
        template_filt  : ndarray
           optimal filtered template in fourier domain


        """

        # channel name 
        channel_name = convert_channel_list_to_name(channels)
        
        # check availability
        if (channel_name not in self._template_filts
            or  template_tag not in  self._template_filts[channel_name]):
            return None
            
        # get array and squeeze
        array = self._template_filts[channel_name][template_tag]
        if squeeze_array:
            array = np.squeeze(array)

        return array



    def template_filt_td(self, channels,
                         template_tag='default',
                         squeeze_array=False):
        """
        (optimal) filtered template, invert fourier transformed

        Get (optimal) filtered template converted back to time domain
        for a specified template tag

        Parameters
        ----------
        channel : str
          channel name

        template_tag : str, optional
          template tag/id
          default: 'default'

        Returns
        -------
        template_filt_td  : 1darray
           optimal filtered template in time domain
        """

        
        # channel name 
        channel_name = convert_channel_list_to_name(channels)
        
        # check availability
        if (channel_name not in self._template_filts_td
            or  template_tag not in  self._template_filts_td[channel_name]):
            raise None
        
            
        # get array and squeeze
        array = self._template_filts_td[channel_name][template_tag]
        if squeeze_array:
            array = np.squeeze(array)

        return array


    # =====================================================
    # Set data
    # =====================================================
    
      
    def add_template(self, channels, templates,
                     template_tag='default',
                     pretrigger_samples=None,
                     maxnorm=False,
                     integralnorm=False,
                     overwrite=False):
        """

        Add NxM template(s) (1D or 2D array) with a user specified tag. 
        Calculate template FFT automatically

        Parameters
        ----------
        channels : str or list of str 
          channel name (s): 
                "chan1" or "chan1|chan2" = ["chan1", "chan2"]

        templates : nd array 1D, 2D, or 3D
           1D: 1x1 template array
           2D: 1xM template array
           3D: NxM template array (N channels, M templates, samples)

        template_tag : str 
           tag name associated to the template
 
        template_group_ids : array
          groups used for multi-template array with each having 2 time degree of freedom
          used for OF calculation (NxMx2 calculation)

        pretrigger_samples : int, optional
            number of pretrigger samples (default: 1/2 trace)

        integralnorm : bool, optional [default = False]
            If set to True, then  template will be normalized
            to an integral of 1, and any optimum filters will
            instead return the optimum integral in units of Coulombs.
            If set to False, then the usual optimum filter amplitudes
            will be returned (in units of Amps).


        Returns
        -------
        None

        """

        # convert channel name
        channel_name = convert_channel_list_to_name(channels)
        channel_list = convert_channel_name_to_list(channels)
        nchans = len(channel_list)

        # convert array to 3D
        if templates.ndim == 1:
            templates = templates[np.newaxis, np.newaxis, :]  # 1 channel, 1 template
        elif templates.ndim == 2:
            templates = templates[np.newaxis, :, :]            # 1 channel, multiple templates
        elif templates.ndim != 3:
            raise ValueError('ERROR: "templates" input must be 1D, 2D, or 3D')        



        # check number of channels
        if (nchans != templates.shape[0]):
            raise ValueError(f'ERROR: Template array has wrong size! '
                             f'Expecting {nchans} channel(s), however '
                             f'template.shape[0] = {templates.shape[0]}')
        
        # check template tag
        if (template_tag is None
            or not isinstance(template_tag, str)):
            raise ValueError('ERROR: "template_tag" expected '
                             'to be a string')
        
        # check if existe already
        if not overwrite:
            if (channel_name in self._templates
                and template_tag in self._templates[channel_name]):
                raise ValueError(
                    f'ERROR: A template with tag "{template_tag}" '
                    f'already exist for channel {channel_name}! Use '
                    f'overwrite=True if needed')
        
        # normalize template
        if maxnorm:
            max_values = np.max(templates, axis=-1, keepdims=True)
            safe_max = np.where(max_values == 0, 1, max_values)
            templates = templates / safe_max
            
                
        # add to dictionary
        if channel_name not in self._templates:
            self._templates[channel_name] = dict()

        # check isnan
        templates[np.isnan(templates)] = 0
        self._templates[channel_name][template_tag] = copy.deepcopy(templates)
        
        # Store number of samples
        # (all data in same OF base object should have same number of samples)
        nbins = templates.shape[-1]
        if  self._nbins is None:
            self._nbins = nbins
        elif nbins != self._nbins:
            raise ValueError(
                f'ERROR: Inconsistent number of samples '
                f'for channel {channel_name}, tag={template_tag}. '
                f'psd/template with same tag must have same '
                f'number of samples!')


        # frequency resolution
        df = self._fs/nbins

        if self._df is None:
            self._df = df

        # FFT
        self._fft_freqs, templates_fft = fft(templates, self._fs, axis=-1)
        if integralnorm:
            dc = templates_fft[...,:1]
            safe_dc = np.where(dc == 0, 1, dc)
            templates_fft = templates_fft / safe_dc

        # store
        if channel_name not in self._templates_fft.keys():
            self._templates_fft[channel_name] = dict()
            
        self._templates_fft[channel_name][template_tag] = templates_fft/nbins

                   
        # store pre-trigger
        if pretrigger_samples is None:
            pretrigger_samples = nbins//2

        if channel_name not in self._pretrigger_samples.keys():
            self._pretrigger_samples[channel_name] = dict()
        self._pretrigger_samples[channel_name][template_tag] =  pretrigger_samples

        # debug
        if self._debug:
            print(f'DEBUG: Add template "{template_tag}" for '
                  f'channel "{channel_name}"!')
            
                
    def set_time_constraints(self, channels, 
                             template_group_ids,
                             fit_window,
                             restrict_time_flag=True,
                             time_constraints_tag='default'):
        """
        Set template time constraints for OF NxMx2
        
        Parameters
        ----------
          
        channels : str or array-like
                array-like or  "|" separated string
                such as "channel1|channel2"

        template_group_ids : 1D numpy array 
             time tage used for multi-template array with each having 
             2 time degree of freedom, Ex [0,1,0,0] -> 1st,3rd, 4th template 
             move together 
             
         fit_window :  used for preparing the time window for nxmx2 filter, 
                   
             time_combinations1 = np.arange(int(fit_window[0][0]), int(fit_window[0][1]))
             time_combinations2 = np.arange(int(fit_window[1][0]), int(fit_window[1][1]))

        restrict_time_flag : boolean, optional

        time_constraints_tag : str
          tag associated to time contraints, to store cosntraint in 
          dictionary

        Return
        ------
        None

        """


        # convert channel name
        channel_name = convert_channel_list_to_name(channels)
        channel_list = convert_channel_name_to_list(channels)
        nchans = len(channel_list)


        # convert fit_window into time_combinations
        time_combinations = self._calc_time_combinations(
            fit_window,
            restrict_time_flag
        )
        

        # create dictionary with all constraints
        constraints_dict = {
            'template_group_ids': template_group_ids,
            'time_combinations': time_combinations,
            'fit_window': fit_window,
            'restrict_time_flag': restrict_time_flag
        }      


        # store time constrains
        if channel_name not in self._time_constraints.keys():
            self._time_constraints[channel_name] = {}
            
        self._time_constraints[channel_name][time_constraints_tag] = (
            constraints_dict
        )


                
    def set_csd(self, channels, csd,
                coupling='AC',
                ignored_frequency_peaks=None,
                ignore_harmonics=False,
                calc_icov=True):
        """
        Add csd, calculate inverse (optional)

        Parameters
        ----------
          
        channels : str or array-like
                array-like or  "|" separated string
                such as "channel1|channel2"
         ORDER in list or string corresponds to matrix channel index

        csd : ndarray
           csd [nchan, nchan, nsamples]
         
        coupling : str, optional [default='AC']
            String that determines if the zero frequency bin of the inverted csd
            should be ignored   when calculating
            the optimum amplitude. If set to 'AC', then the zero
            frequency bin is ignored. If set to anything else, then the
            zero frequency bin is kept.

        ignored_frequency_peaks : float or list of float [default=None] 
            frequencies that are ignored in optimal filter
            calculation. The nearest bin (both positive/negative frequencies) from
            given frequency is selected and CSD frequency set to infinity.
         
        ignore_harmonics : boolean
            if True, harmonics of the ignored_frequency_peaks are also ignored up to max frequency
      
        Returns
        -------
        None

        """
        
        # convert to string if needed (if str, return same string)
        channel_name = convert_channel_list_to_name(channels)

        # convert to 3D if needed
        if csd.ndim == 1:
            csd = csd[np.newaxis, np.newaxis, :]
            
        
        # check if same length as template
        nbins = csd.shape[-1]
     
        if self._nbins is None:
            self._nbins = nbins
        elif nbins != self._nbins:
            raise ValueError(
                f'ERROR: Inconsistent number of samples '
                f'for channel {channels}. '
                f'csd (# samples={nbins}) and template '
                f'(# samples={self._nbins}) must have same '
                f'number of samples!')

        # frequency vectror
        self._fft_freqs = fftfreq(nbins, self._fs)
        
        # add to dictionary
        self._csd[channel_name] = csd

        # frequency resolution
        if self._df is None:
            self._df= self._fs/nbins
            
        # calculate icov
        if calc_icov:
            self.calc_icovf(channel_name,
                            coupling=coupling,
                            ignored_frequency_peaks=ignored_frequency_peaks,
                            ignore_harmonics=ignore_harmonics)
            
            
    def set_psd(self, channel, psd, coupling='AC',
                ignored_frequency_peaks=None,
                ignore_harmonics=False):
        """
        Add psd for specified channel
        If psd already exist, it is overwritten


        Parameters
        ----------
        channel : str
          channel name

        psd : ndarray
           psd 1d array

        coupling : str, optional [default='AC']
            String that determines if the zero frequency bin of the psd
            should be ignored (i.e. set to infinity) when calculating
            the optimum amplitude. If set to 'AC', then ths zero
            frequency bin is ignored. If set to anything else, then the
            zero frequency bin is kept.

        ignored_frequency_peaks : float or list of float [default=None] 
            frequencies that are ignored in optimal filter
            calculation. The nearest bin (both positive/negative frequencies) from
            given frequency is selected and CSD frequency set to infinity.
              
        ignore_harmonics : boolean
            if True, harmonics of the ignored_frequency_peaks are also ignored up to max frequency

        Returns
        -------
        None

        """

        # check coupling
        if coupling=="AC":
            psd[0] = np.inf

        # check if same length as template
        nbins = psd.shape[-1]

        if self._nbins is None:
            self._nbins = nbins
        elif nbins != self._nbins:
            raise ValueError(
                f'ERROR: Inconsistent number of samples '
                f'for channel {channel} '
                f'psd/template with same tag must have same '
                f'number of samples!')


        # frequency vectror
        self._fft_freqs = fftfreq(nbins, self._fs)
        
        # add to dictionary
        csd = psd
        if psd.ndim == 1:
            csd = psd[np.newaxis, np.newaxis, :]  # now shape is (1, 1, nfreq)

        self.set_csd(channel, csd, coupling=coupling,
                     ignored_frequency_peaks=ignored_frequency_peaks,
                     ignore_harmonics=ignore_harmonics)
        


    def calc_icovf(self, channels,
                   coupling='AC',
                   ignored_frequency_peaks=None,
                   ignore_harmonics=False):
        """
        A function that inverts the csd or covariance noise matrix between channels. 
        
        Parameters
        ----------
        
        channels : str or list
            multiple channels as a single "|" separated string
            such as "channel1|channel2"

        coupling : str, optional [default='AC']
            String that determines if the zero frequency bin of the csd
            should be ignored   when calculating
            the optimum amplitude. If set to 'AC', then the zero
            frequency bin is ignored. If set to anything else, then the
            zero frequency bin is kept.

        ignored_frequency_peaks : float or list of float [default=None] 
            frequencies that are ignored in optimal filter
            calculation. The nearest bin (both positive/negative frequencies) from
            given frequency is selected and CSD frequency set to infinity.

        ignore_harmonics : boolean
            if True, harmonics of the ignored_frequency_peaks are also ignored up to max frequency

        Returns
        -------
        
        None
        """
        
        # convert to name/list
        channel_name = convert_channel_list_to_name(channels)
        
        #I should add lines that make sure csd is instantiated first
        covf = np.copy(self.csd(channel_name)) #an ndarray for a combination of channels
        covf *= self._df #[A^2/Hz] -> [A^2]

        temp_icovf = np.zeros_like(covf, dtype='complex')
        for ii in range(self._nbins):
            temp_icovf[:,:,ii] = pinv(covf[:,:,ii]) #1/A^2

        if coupling == 'AC':
            temp_icovf[:,:,0] = 0.0


        if ignored_frequency_peaks is not None:

            # normalize input to a list
            if not isinstance(ignored_frequency_peaks, list):
                ignored_frequency_peaks = [ignored_frequency_peaks]
                print(ignored_frequency_peaks)
            freqs = np.asarray(self._fft_freqs)
            fmax = float(np.max(np.abs(freqs)))
            
            # tolerance ?
            """
            # estimate frequency-bin spacing (robust to sign order)
            # sort by absolute frequency to avoid the 0 jump
            sort_idx = np.argsort(np.abs(freqs))
            df_est = np.median(np.diff(np.sort(np.abs(freqs))))
            if not np.isfinite(df_est) or df_est <= 0:
                # fallback if something weird with freqs
                df_est = np.min(np.abs(np.diff(np.sort(freqs)))) if freqs.size > 1 else 0.0

            # tolerance: half a bin by default
            tol = 0.5 * df_est if df_est > 0 else 0.0
            """

            # accumulate all bins to zero in one mask
            mask_all = np.zeros_like(freqs, dtype=bool)

            for f in ignored_frequency_peaks:
                if f is None or not np.isfinite(f):
                    continue

                f = float(np.abs(f))
                if f == 0.0:
                    # Special case: DC (closest-to-zero bin)
                    idx0 = int(np.argmin(np.abs(freqs)))
                    mask_all[idx0] = True
                    continue

                # Build the target frequencies to zero
                if ignore_harmonics:
                    kmax = int(np.floor(fmax / f))
                    if kmax <= 0:
                        continue
                    targets = f * np.arange(1, kmax + 1, dtype=float)
                else:
                    targets = np.array([f], dtype=float)
                
                # For each target harmonic, hit both + and - sides
                for t in targets:
                    
                    # positive side
                    idx_pos = int(np.argmin(np.abs(freqs - t)))
                    #if np.abs(freqs[idx_pos] - t) <= tol:
                    mask_all[idx_pos] = True

                    # negative side
                    idx_neg = int(np.argmin(np.abs(freqs + t)))
                    #if np.abs(freqs[idx_neg] + t) <= tol:
                    mask_all[idx_neg] = True
                        
            # apply once
            temp_icovf[..., mask_all] = 0.0


        self._icovf[channel_name] = temp_icovf
        
        
    def clear_signal(self):
        """
        Method to intialize calculated signal
        parameters

        Parameters
        ----------
           
        channels : str or array-like
           if multi_channels:
                array-like or  "|" separated string
                such as "channel1|channel2"

        Return
        ---------
        None

        """
                
        # signal
        self._signals = dict()
        self._signals_fft = dict()
        
        # (optimal) filtered  signals and templates
        self._signals_filts = dict()
        self._signals_filts_td = dict()
        self._template_filts = dict()
        self._template_filts_td = dict()
        
                   
    def update_signal(self, channels, signals,
                      calc_fft=True):
              
        """
        Method to update new signal trace for single channel, 
        needs to be called each event

        Parameters
        ----------
        channel : str or list of ste
          channel name

        signal : ndarray
           the signal that we want to apply the optimum filter to
           (units should be Amps). Dimensions need to match number of 
           signals

        calc_fft : bool, optional
           If true calculate signal fft
           Default: True

        Return
        ------
        None
        """

        # let's get channel list
        channel_list = convert_channel_name_to_list(channels)
        channel_name = convert_channel_list_to_name(channels)
        nchans = len(channel_list)

        # check signal size
        if signals.ndim == 1:
            signals = signals[np.newaxis, :]
       
        nchans_array = signals.shape[0]
        nbins_array = signals.shape[-1]

        if (nchans_array != nchans):
            raise ValueError(f'ERROR: number of channels ({nchans}) '
                             f'does not match signal array ({nchans_array})')

        if self._nbins is None:
            self._nbins = nbins_array
        elif (nbins_array != self._nbins):
            raise ValueError(f'ERROR:Inconsistent number of samples '
                             f'between signal ({nbins_array}) '
                             f'and template/psd ({ self._nbins})')
        
             
        # loop channels and add signal
        for ichan, chan in enumerate(channel_list):
            
            signal =  signals[ichan,:]
            
            # debug
            if self._debug:
                print(f'DEBUG: Update signal for channel '
                      f'"{chan}"!')
                
            # check if already updated
            if chan in self._signals:
                if self._debug:
                    print(f'WARNING: A signal already exist for '
                          f'channel {chan}.')
                    continue
                          
            # add signal
            self._signals[chan] = signal
            
            # add fft ("forward" normalization)
            if calc_fft:
                f, signal_fft = fft(signal, self._fs, axis=-1)
                self._signals_fft[chan] = signal_fft/self._nbins

                if self._fft_freqs is None:
                    self._fft_freqs = f
                elif (not np.array_equal(self._fft_freqs,f)):
                    raise ValueError(f'ERROR:Inconsistent frequency array '
                                     f'between signal and template/psd')

                
    def calc_phi(self, channels, template_tag=None,
                 calc_weight=True):
        """

        Calculates NxM optimal filter matrix across the specified channels 
        and templates. Depends on the templates and the inverted covariance matrix. 
        This function also checks that precalculations are covered. 
        
        For 1x1: phi = template_fft* / psd

        Parameters
        ----------
        channels : str or list of str
          channel name (s)

        template_tag : str  (optional)
           tag associated with template
           if None, calculate for tags

        Return
        ------
        None
        """

        # convert to name
        channel_name = convert_channel_list_to_name(channels)
        
        # check if noise available
        if channel_name not in self._csd:
            raise ValueError(f'ERROR: Noise csd not availabe for '
                             f'channel {channel_name} !')

        
        # get noise data (inverte covariance matrix)
        if self.icovf(channel_name) is None:
            raise ValueError(f'ERROR: Inverted CSD has not been calculated '
                             f'for channel {channel_name} !')

        icovf = copy.deepcopy(self.icovf(channel_name))


        # get list of tags
        template_tags_list = [template_tag]
        if template_tag is None:
            template_tags_list = self.template_tags(channel_name)

        for tag in template_tags_list:

            # get template FFT
            if (channel_name not in self._templates_fft
                or tag not in self._templates_fft[channel_name]):
                raise ValueError(f'ERROR: Missing template fft for '
                                 f'channel {channel_name}, '
                                 f'template tag = {tag}')

            template_fft = self._templates_fft[channel_name][tag]

            # calculate optimal filter
            
            # Using Einstein summation:
            # - Let k be the channel index,
            # - j be the template index,
            # - and n the frequency bin.
            
            phi = np.einsum('kjn,kin->ijn', np.conjugate(template_fft), icovf)

        
            # save
            if channel_name not in self._phis:
                self._phis[channel_name] = dict()

            self._phis[channel_name][tag] =  copy.deepcopy(phi)


            # calc weight matrix
            if calc_weight:
                self.calc_weight(channel_name, template_tag=tag)
                


    
        
    def calc_weight(self, channels, template_tag=None):
        """
        A function that calculates the inverted and non inverted weighting (or norm) matrix. 
        Depends on the optimal filter matrix (phi) and the template matrix.
        This function also checks that the phi matrix has been precomputed. 
        # 1D  self._norms[channel][tag] = (
                np.real(np.dot(self._phis[channel][tag],
                               self._templates_fft[channel][tag]))*self._df
            )
        
        Parameters
        ----------
        
        channels : list or string
            channel names

        template_tags: 2D array (optional)
           array of template tags
        
        Returns
        -------
        
        None
        """

        # convert to name
        channel_name = convert_channel_list_to_name(channels)
    

        template_tags_list = [template_tag]
        if template_tag is None:
            template_tags_list = self.template_tags(channel_name)
      
        for tag in template_tags_list:
        
            # get phi
            if self.phi(channel_name, tag) is None:
                self.calc_phi(channel_name, tag)

            phi = self.phi(channel_name, tag,
                           squeeze_array=False)
        
            # get template FFT
            template_fft = self.template_fft(channel_name, tag,
                                             squeeze_array=False)
            
            # Compute the weight matrix
            temp_w = np.einsum('cif,cjf->ij', phi, template_fft)
            
            # take the pseudoinverse of its real part,
            weight =  copy.deepcopy(np.real(temp_w))
            
            # store
            if channel_name not in self._weights:
                self._weights[channel_name] = dict()
                self._iweights[channel_name] = dict()
                
            self._weights[channel_name][tag] =  weight
            self._iweights[channel_name][tag] = pinv(weight)
            

            
    def calc_p_matrix(self, channels,
                      template_tag=None,
                      time_constraints_tag=None):
        """
        P matrix calculation for NxMx2 for specified template tag
        and time constraints tag. If tags are None, calculate with 
        all available tags.
        """

        # convert to name/list
        channel_list = convert_channel_name_to_list(channels)
        channel_name = convert_channel_list_to_name(channels)
        nchans = len(channel_list)
        
        # template tags
        template_tag_list = [template_tag]
        if template_tag is None:
            template_tag_list = self.template_tags(channel_name)
            
        if (len(template_tag_list) == 0):
            raise ValueError(f'ERROR: No template tag available for '
                             f'channel {channel_name} !')
        
        # time constraint tag
        time_constraints_tag_list = [time_constraints_tag]
        if (len(time_constraints_tag_list) == 0):
            raise ValueError(f'ERROR: No time constraint tag available for '
                             f'channel {channel_name} !')
        


        # loop template and time constraint tags
        for tag in template_tag_list:

            # get template FFT
            template_fft = self.template_fft(channel_name,
                                             template_tag=tag,
                                             squeeze_array=False)

            # number of template
            ntmps = template_fft.shape[1]

            
            # get optimal filter
            phi = self.phi(channel_name,
                           template_tag=tag,
                           squeeze_array=False)

                       
            if phi is None:
                self.calc_phi(channel_name, template_tag=tag)
                phi = self.phi(channel_name,
                               template_tag=tag,
                               squeeze_array=False)
                      
            # loop time constraints
            for constraints_tag in time_constraints_tag_list:


                # check if p_matrix already calculated
                p_matrix = self.p_matrix(
                    channel_name,
                    template_tag=tag,
                    time_constraints_tag=constraints_tag)

                if p_matrix is not None:
                    continue
                

                # get constraints
                template_group_ids = self.template_group_ids(
                    channel_name,
                    time_constraints_tag=constraints_tag)

                t0s = self.time_combinations(
                    channel_name,
                    time_constraints_tag=constraints_tag
                )

                # group id matrix
                time_diff  = template_group_ids[:, None] - template_group_ids[None, :]
            
                
                p = np.zeros((self._nbins, ntmps, ntmps ), dtype=np.complex128)
                np.einsum('jii->ji', p)[:] = 1
                for itmp in range(ntmps):
                    for jtmp in range(ntmps):
                        sum = 0.0 + 0.0j
                        for jchan in range(nchans):
                            if (time_diff[itmp, jtmp] != 0):
                                sum += np.fft.ifft(
                                    phi[jchan,itmp, :]\
                                    * template_fft[jchan,jtmp,:]
                                )*self._nbins
                            if (time_diff[itmp,jtmp] == 0):
                                sum += np.sum(
                                    phi[jchan,itmp,:]\
                                    * template_fft[jchan,jtmp,:]
                                )
                                
                        if (time_diff[itmp, jtmp] == -1):
                            p[:,itmp,jtmp] = p[:,jtmp,itmp] = np.real(sum)
                        elif (time_diff[itmp, jtmp] == 0):
                            p[:,itmp,jtmp] = p[:,jtmp,itmp] = np.real(sum)

                p_inv = np.linalg.pinv(p)

                # add constraint
                p_matrix =  np.zeros((t0s[:,0].shape[0], ntmps, ntmps))
                p_matrix_inv =  np.zeros((t0s[:,0].shape[0], ntmps, ntmps))
                
                np.einsum('jii->ji', p_matrix_inv)[:] = 1

                for itmps in range(ntmps):
                    for jtmps in range(ntmps):
                        p_matrix[:, itmps, jtmps] = (
                            p[t0s[:,0]-t0s[:,1]][:, itmps, jtmps]
                        )
                        p_matrix_inv[:, itmps, jtmps] = (
                            p_inv[t0s[:,0]- t0s[:,1]][:, itmps, jtmps]
                        )

                        
                # save
                combined_tag = f'{tag}_{constraints_tag}'
                
                if channel_name not in self._p_matrix:
                    self._p_matrix[channel_name] = dict()
                    self._p_matrix_inv[channel_name] = dict()
                    
                self._p_matrix[channel_name][combined_tag] = (
                     copy.deepcopy(p_matrix)
                )
                    
                self._p_matrix_inv[channel_name][combined_tag] = (
                     copy.deepcopy(p_matrix_inv)
                )
                
                
  
    def calc_signal_filt(self, channels, template_tag=None):
        """
        A function that calculates the filtered signal matrix in frequency domain.
        That is, the signal matrix multiplied by the optimal filter matrix (phi). 
        
        Parameters
        ----------
        
        channels : str or list
            multiple channels as a list or 
            as "|" separated string such as "channel1|channel2"

        template_tag : str (optional)
          template tag, if None: all available
        
        Returns
        -------
        
        None
        """
        
        # convert to name/list
        channel_name = convert_channel_list_to_name(channels)

        # get list of tags
        template_tags_list = [template_tag]
        if template_tag is None:
            template_tags_list = self.template_tags(channel_name)

        for tag in template_tags_list:

            
            # get optimal filter
            phi = self.phi(channel_name,
                           template_tag=tag,
                           squeeze_array=False)
            
            if phi is None:
                self.calc_phi(channel_name, template_tag=tag)
                phi = self.phi(channel_name,
                               template_tag=tag,
                               squeeze_array=False)
                
                       
            signal_fft = self.signal_fft(channel_name)
             
            # calculate
            filtered_signal = np.einsum('cif,cf->if', phi, signal_fft)
            
            # save
            if channel_name not in self._signals_filts:
                self._signals_filts[channel_name] = dict()
            self._signals_filts[channel_name][tag] =  copy.deepcopy(
                filtered_signal
            )


    def calc_signal_filt_td(self, channels,
                            template_tag=None):
        """
        A function that calculates the optimally filtered signal matrix in time domain.
        In other words, the ifft of the filtered signal matrix. 
         
        Parameters
        ----------
        
        channels : str or list 
            multiple channels as a single "|" separated string
            such as "channel1|channel2"
        
        Returns
        -------
        
        None
        """

        # channel name
        channel_name = convert_channel_list_to_name(channels)
        
        # get list of tags
        template_tags_list = [template_tag]
        if template_tag is None:
            template_tags_list = self.template_tags(channel_name)

        for tag in template_tags_list:
            
            # check if calculated already
            signal_td = self.signal_filt_td(
                channel_name,
                template_tag=tag
            )
            
            if signal_td  is not None:
                continue
                

            # check filtered signal has been calculated
            filtered_signal = self.signal_filt(channel_name,
                                             template_tag=tag)

            if filtered_signal is None:
                self.calc_signal_filt(channel_name,
                                      template_tag=tag)

                filtered_signal = self.signal_filt(channel_name,
                                                   template_tag=tag)
                
            filtered_signal_td = np.real(ifft(filtered_signal*self._nbins))

            # store
            if channel_name not in self._signals_filts_td:
                self._signals_filts_td[channel_name] = dict()
            
            self._signals_filts_td[channel_name][tag] = (
                copy.deepcopy(filtered_signal_td)
            )

      
    def _calc_time_combinations(self,
                                fit_window,
                                restrict_time_flag):
        """
        calc time combinations OF NxMx2
        
        Parameters
        ----------
            
        fit_window :   array-likein integer
           used for preparing the time window for nxmx2 fit
           in samples. Ex:  [[-625,625],[-625,1250]]

        
        Return
        ------
        time combinations

        """
                
        # time constraints
        time_combinations1 = None
        time_combinations2 = None
        if fit_window is None:
            time_combinations1 = np.arange(int(-self._nbins/2),
                                           int(self._nbins/2))
            time_combinations2 = np.arange(int(-self._nbins/2),
                                           int(self._nbins/2))
        else:
            time_combinations1 = np.arange(int(fit_window[0][0]),
                                           int(fit_window[0][1]))
            time_combinations2 = np.arange(int(fit_window[1][0]),
                                           int(fit_window[1][1]))

        time_combinations = None
        if restrict_time_flag:
            X,Y = np.meshgrid(time_combinations1, time_combinations2)
            mask = X <= Y
            indices = np.where(mask)
            time_combinations = np.column_stack(( X[indices] ,Y[indices] ))
        else:
            time_combinations = (
                np.stack(np.meshgrid(time_combinations1,
                                     time_combinations2), -1).reshape(-1, 2)
            )

        return time_combinations



